#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the model and editor of the `acct_gather.conf` configuration file."""


from constants import EXAMPLE_ACCT_GATHER_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import AcctGatherConfig, ModelError, acctgatherconfig

EXPECTED_ACCT_GATHER_DUMPS_OUTPUT = """
energyipmifrequency=1
energyipmicalcadjustment=yes
energyipmipowersensors=node=16,19;socket1=19,26;knc=16,19
energyipmiusername=testipmiusername
energyipmipassword=testipmipassword
energyipmitimeout=10
profilehdf5dir=/mydir
profilehdf5default=all
profileinfluxdbdatabase=acct_gather_db
profileinfluxdbdefault=all
profileinfluxdbhost=testhostname
profileinfluxdbpass=testpassword
profileinfluxdbrtpolicy=testpolicy
profileinfluxdbuser=testuser
profileinfluxdbtimeout=10
infinibandofedport=0
sysfsinterfaces=enp0s1
""".strip()


class TestAcctGatherConfigEditor(TestCase):
    """Unit tests for the model and editor of the `acct_gather.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/acct_gather.conf", contents=EXAMPLE_ACCT_GATHER_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of the `acct_gather.conf` model editor."""
        config = acctgatherconfig.loads(EXAMPLE_ACCT_GATHER_CONFIG)
        self.assertEqual(config.energy_ipmi_frequency, 1)
        self.assertTrue(config.energy_ipmi_calc_adjustment)
        self.assertDictEqual(
            config.energy_ipmi_power_sensors,
            {"node": [16, 19], "socket1": [19, 26], "knc": [16, 19]},
        )
        self.assertEqual(config.energy_ipmi_username, "testipmiusername")
        self.assertEqual(config.energy_ipmi_password, "testipmipassword")
        self.assertEqual(config.energy_ipmi_timeout, 10)
        self.assertEqual(config.profile_hdf5_dir, "/mydir")
        self.assertListEqual(config.profile_hdf5_default, ["all"])
        self.assertEqual(config.profile_influxdb_database, "acct_gather_db")
        self.assertListEqual(config.profile_influxdb_default, ["all"])
        self.assertEqual(config.profile_influxdb_host, "testhostname")
        self.assertEqual(config.profile_influxdb_pass, "testpassword")
        self.assertEqual(config.profile_influxdb_rt_policy, "testpolicy")
        self.assertEqual(config.profile_influxdb_user, "testuser")
        self.assertEqual(config.profile_influxdb_timeout, 10)
        self.assertEqual(config.infiniband_ofed_port, 0)
        self.assertListEqual(config.sysfs_interfaces, ["enp0s1"])

    def test_dumps(self) -> None:
        """Test `dumps` method of the `acct_gather.conf` model editor."""
        config = acctgatherconfig.loads(EXAMPLE_ACCT_GATHER_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(acctgatherconfig.dumps(config), EXPECTED_ACCT_GATHER_DUMPS_OUTPUT)
        # New config and old config should not be equal since the editor strips all comments.
        self.assertNotEqual(acctgatherconfig.dumps(config), EXAMPLE_ACCT_GATHER_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` context manager from the acctgatherconfig module."""
        with acctgatherconfig.edit("/etc/slurm/acct_gather.conf") as config:
            config.energy_ipmi_frequency = 2
            config.energy_ipmi_calc_adjustment = False
            config.energy_ipmi_power_sensors = {
                "node": [20, 21],
                "socket1": [10, 26],
                "knc": [16, 19],
            }
            config.energy_ipmi_username = "testipmiusername1"
            config.energy_ipmi_password = "testipmipassword1"
            config.energy_ipmi_timeout = 20
            config.profile_hdf5_dir = "/mydir1234"
            config.profile_hdf5_default = ["none"]
            config.profile_influxdb_database = "test_acct_gather_db"
            config.profile_influxdb_default = ["none"]
            config.profile_influxdb_host = "testhostname1"
            config.profile_influxdb_pass = "testpassword1"
            config.profile_influxdb_rt_policy = "testpolicy1"
            config.profile_influxdb_user = "testuser1"
            config.profile_influxdb_timeout = 20
            config.infiniband_ofed_port = 1
            config.sysfs_interfaces = ["enp0s2"]

        config = acctgatherconfig.load("/etc/slurm/acct_gather.conf")
        self.assertEqual(config.energy_ipmi_frequency, 2)
        self.assertFalse(config.energy_ipmi_calc_adjustment)
        self.assertDictEqual(
            config.energy_ipmi_power_sensors,
            {"node": [20, 21], "socket1": [10, 26], "knc": [16, 19]},
        ),
        self.assertEqual(config.energy_ipmi_username, "testipmiusername1")
        self.assertEqual(config.energy_ipmi_password, "testipmipassword1")
        self.assertEqual(config.energy_ipmi_timeout, 20)
        self.assertEqual(config.profile_hdf5_dir, "/mydir1234")
        self.assertEqual(config.profile_hdf5_default, ["none"])
        self.assertEqual(config.profile_influxdb_database, "test_acct_gather_db")
        self.assertEqual(config.profile_influxdb_default, ["none"])
        self.assertEqual(config.profile_influxdb_host, "testhostname1")
        self.assertEqual(config.profile_influxdb_pass, "testpassword1")
        self.assertEqual(config.profile_influxdb_rt_policy, "testpolicy1")
        self.assertEqual(config.profile_influxdb_user, "testuser1")
        self.assertEqual(config.profile_influxdb_timeout, 20)
        self.assertEqual(config.infiniband_ofed_port, 1)
        self.assertEqual(config.sysfs_interfaces, ["enp0s2"])

    def test_load_fail(self) -> None:
        """Test that bogus values are automatically caught when loading a new model."""
        # Catch if field value is unexpected. e.g. not a boolean value.
        with self.assertRaises(ValueError):
            AcctGatherConfig.from_str(
                """
                energyipmicalcadjustment=totally
                """
            )

        # Catch if field is unexpected or in the incorrect syntax.
        with self.assertRaises(ModelError):
            AcctGatherConfig.from_str(
                """
                observe_the_jobs=totally
                """
            )

        # Catch if field value is invalid when checking against the model schema.
        with self.assertRaises(ModelError):
            AcctGatherConfig.from_str(
                """
                EnergyIPMIFrequency=ALL THE TIME!!
                """
            )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = acctgatherconfig.loads(EXAMPLE_ACCT_GATHER_CONFIG)
            config.profile_influxdbrt_policy = "newtestpolicy"

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = acctgatherconfig.loads(EXAMPLE_ACCT_GATHER_CONFIG)
            _ = config.profile_influxdbrt_policy



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Models representing the `slurm.conf` configuration file."""

__all__ = [
    "DownNodes",
    "DownNodesList",
    "FrontendNode",
    "FrontendNodeMapping",
    "Node",
    "NodeMapping",
    "NodeSet",
    "NodeSetMapping",
    "Partition",
    "PartitionMapping",
    "SlurmConfig",
    "SlurmConfigEditor",
]

from collections.abc import Callable, Iterable
from typing import Annotated, Any, cast

from .core.base import (
    Metadata,
    Mode,
    Model,
    ModelList,
    ModelMapping,
    classproperty,
    make_model_builder,
)
from .core.callback import (
    ColonSepCallback,
    CommaDictCallback,
    CommaDictColonArrayCallback,
    CommaDictColonPairCallback,
    CommaSepCallback,
    IntBoolCallback,
    MultilineCallback,
    QuoteCallback,
    StrBoolCallback,
    make_callback,
)
from .core.editor import BaseEditor
from .core.schema import (
    DOWN_NODES_LIST_MODEL_SCHEMA,
    DOWN_NODES_MODEL_SCHEMA,
    FRONTEND_NODE_MAPPING_MODEL_SCHEMA,
    FRONTEND_NODE_MODEL_SCHEMA,
    NODE_MAPPING_MODEL_SCHEMA,
    NODE_MODEL_SCHEMA,
    NODESET_MAPPING_MODEL_SCHEMA,
    NODESET_MODEL_SCHEMA,
    PARTITION_MAPPING_MODEL_SCHEMA,
    PARTITION_MODEL_SCHEMA,
    SLURM_CONFIG_MODEL_SCHEMA,
)


def _mcs_parameters_parser(value: str) -> dict[str, bool | list[str]]:
    params = value.split(":", maxsplit=1)
    result: dict[str, bool | list[str]] = dict.fromkeys(params[0].split(","), True)
    try:
        plugin_params = params[1]
        result["mcs_plugin_parameters"] = plugin_params.split("|")
    except IndexError:
        pass

    return result


def _mcs_parameters_marshaller(value: dict[str, bool | Iterable[str]]) -> str:
    plugin_params = "|".join(cast(Iterable[str], value.pop("mcs_plugin_parameters", [])))
    params = ",".join(k for k in value if value[k])
    if plugin_params:
        params += f":{plugin_params}"

    return params


MCSParametersCallback = make_callback(_mcs_parameters_parser, _mcs_parameters_marshaller)


class DownNodes(Model):
    """Model representing declared down nodes in the `slurm.conf` configuration file."""

    down_nodes: Annotated[list[str] | None, Metadata(primary=True, callback=CommaSepCallback)]
    reason: Annotated[str | None, Metadata(callback=QuoteCallback)]
    state: str | None

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return DOWN_NODES_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        return Mode.ONELINE


class FrontendNode(Model):
    """Model representing a frontend node in the `slurm.conf` configuration file."""

    frontend_name: Annotated[str | None, Metadata(primary=True)]
    frontend_addr: str | None
    allow_groups: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    allow_users: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    deny_groups: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    deny_users: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    port: int | None
    reason: Annotated[str | None, Metadata(callback=QuoteCallback)]
    state: str | None

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return FRONTEND_NODE_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        return Mode.ONELINE


class Node(Model):
    """Model representing a node in the `slurm.conf` configuration file."""

    node_name: Annotated[str | None, Metadata(primary=True)]
    node_hostname: str | None
    node_addr: str | None
    bcast_addr: str | None
    boards: int | None
    core_spec_count: int | None
    cores_per_socket: int | None
    cpu_bind: str | None
    cpus: int | None
    cpu_spec_list: Annotated[list[int] | None, Metadata(callback=CommaSepCallback)]
    features: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    gres: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    mem_spec_limit: int | None
    port: int | None
    procs: int | None
    real_memory: int | None
    reason: Annotated[str | None, Metadata(callback=QuoteCallback)]
    sockets: int | None
    sockets_per_board: int | None
    state: str | None
    threads_per_core: int | None
    tmp_disk: int | None
    weight: int | None

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return NODE_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        return Mode.ONELINE


class NodeSet(Model):
    """Model representing a node set in the `slurm.conf` configuration file."""

    node_set: Annotated[str | None, Metadata(primary=True)]
    feature: str | None
    nodes: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return NODESET_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        return Mode.ONELINE


class Partition(Model):
    """Model representing a partition in the `slurm.conf` configuration file."""

    partition_name: Annotated[str | None, Metadata(primary=True)]
    alloc_nodes: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    allow_accounts: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    allow_groups: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    allow_qos: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    alternate: str | None
    cpu_bind: str | None
    default: bool | None
    default_time: int | str | None
    def_cpu_per_gpu: int | None
    def_mem_per_cpu: int | None
    def_mem_per_gpu: int | None
    def_mem_per_node: int | None
    deny_accounts: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    deny_qos: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    disable_root_jobs: bool | None
    exclusive_topo: bool | None
    exclusive_user: bool | None
    grace_time: int | None
    hidden: bool | None
    lln: bool | None
    max_cpus_per_node: int | None
    max_cpus_per_socket: int | None
    max_mem_per_cpu: int | None
    max_mem_per_node: int | None
    max_nodes: int | str | None
    max_time: int | str | None
    min_nodes: int | None
    nodes: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    over_subscribe: str | None
    over_time_limit: int | str | None
    power_down_on_idle: bool | None
    preempt_mode: str | None
    priority_job_factor: int | None
    priority_tier: int | None
    qos: str | None
    req_resv: bool | None
    resume_timeout: int | None
    root_only: bool | None
    select_type_parameters: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    state: str | None
    suspend_time: int | str | None
    suspend_timeout: int | None
    tres_billing_weights: Annotated[dict[str, str] | None, Metadata(callback=CommaDictCallback)]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return PARTITION_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        return Mode.ONELINE


class DownNodesList(ModelList[DownNodes]):
    """List of down nodes entries in the `slurm.conf` configuration file."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return DOWN_NODES_LIST_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _down_nodes_list_model_builder


class FrontendNodeMapping(ModelMapping[str, FrontendNode]):
    """Mapping of frontend node names to corresponding frontend nodes."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return FRONTEND_NODE_MAPPING_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _frontend_node_mapping_model_builder


class NodeMapping(ModelMapping[str, Node]):
    """Mapping of node names to corresponding nodes."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return NODE_MAPPING_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _node_mapping_model_builder


class NodeSetMapping(ModelMapping[str, NodeSet]):
    """Mapping of node set names to corresponding node sets."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return NODESET_MAPPING_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _nodeset_mapping_model_builder


class PartitionMapping(ModelMapping[str, Partition]):
    """Mapping of partition names to corresponding partitions."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return PARTITION_MAPPING_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _partition_mapping_model_builder


class SlurmConfig(Model):
    """Model representing the `slurm.conf` configuration file."""

    accounting_storage_backup_host: str | None
    accounting_storage_enforce: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    accounting_storage_external_host: str | None
    accounting_storage_host: str | None
    accounting_storage_parameters: Annotated[
        dict[str, str | list[str]] | None,
        Metadata(callback=CommaDictCallback),
    ]
    accounting_storage_pass: str | None
    accounting_storage_port: int | None
    accounting_storage_tres: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    accounting_storage_type: str | None
    accounting_storage_user: str | None
    accounting_store_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    acct_gather_node_freq: int | None
    acct_gather_energy_type: str | None
    acct_gather_interconnect_type: str | None
    acct_gather_filesystem_type: str | None
    acct_gather_profile_type: str | None
    allow_spec_resources_usage: bool | None
    auth_alt_types: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    auth_alt_parameters: Annotated[
        dict[str, bool | int | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    auth_info: Annotated[dict[str, bool | int | str] | None, Metadata(callback=CommaDictCallback)]
    auth_type: str | None
    batch_start_timeout: int | None
    bcast_exclude: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    bcast_parameters: Annotated[
        dict[str, bool | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    burst_buffer_type: str | None
    cli_filter_plugins: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    cluster_name: str | None
    communication_parameters: Annotated[
        dict[str, bool | int] | None, Metadata(callback=CommaDictCallback)
    ]
    complete_wait: int | None
    cpu_freq_def: str | None
    cpu_freq_governors: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    cred_type: str | None
    debug_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    def_cpu_per_gpu: int | None
    def_mem_per_cpu: int | None
    def_mem_per_gpu: int | None
    def_mem_per_node: int | None
    dependency_parameters: Annotated[
        dict[str, bool | int] | None,
        Metadata(callback=CommaDictCallback),
    ]
    disable_root_jobs: bool | None
    eio_timeout: int | None
    enforce_part_limits: str | None
    epilog: str | None
    epilog_msg_time: int | None
    epilog_slurmctld: str | None
    fair_share_dampening_factor: int | None
    federation_parameters: Annotated[
        dict[str, bool] | None,
        Metadata(callback=CommaDictCallback),
    ]
    first_job_id: int | None
    get_env_timeout: int | None
    gres_types: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    group_update_force: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    group_update_time: int | None
    gpu_freq_def: int | str | None
    health_check_interval: int | None
    health_check_node_state: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    health_check_program: str | None
    inactive_limit: int | None
    include: Annotated[list[str] | None, Metadata(sep=" ", unique=False, callback=MultilineCallback)]
    interactive_step_options: str | None
    job_acct_gather_type: str | None
    job_acct_gather_frequency: Annotated[
        dict[str, int] | None,
        Metadata(callback=CommaDictCallback),
    ]
    job_acct_gather_params: Annotated[
        dict[str, bool] | None,
        Metadata(callback=CommaDictCallback),
    ]
    job_comp_host: str | None
    job_comp_loc: str | None
    job_comp_params: Annotated[
        dict[str, bool | int | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    job_comp_pass: str | None
    job_comp_port: int | None
    job_comp_type: str | None
    job_comp_user: str | None
    job_container_type: str | None
    job_file_append: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    job_requeue: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    job_submit_plugins: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    kill_on_bad_exit: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    kill_wait: int | None
    max_batch_requeue: int | None
    node_features_plugins: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    launch_parameters: Annotated[
        dict[str, bool] | None,
        Metadata(callback=CommaDictCallback),
    ]
    licenses: Annotated[
        dict[str, bool | int] | None,
        Metadata(callback=CommaDictColonPairCallback),
    ]
    log_time_format: str | None
    mail_domain: str | None
    mail_prog: str | None
    max_array_size: int | None
    max_dbd_msgs: int | None
    max_job_count: int | None
    max_job_id: int | None
    max_mem_per_cpu: int | None
    max_mem_per_node: int | None
    max_node_count: int | None
    max_step_count: int | None
    max_tasks_per_node: int | None
    mcs_parameters: Annotated[
        dict[str, bool | list[str]] | None,
        Metadata(callback=MCSParametersCallback),
    ]
    mcs_plugin: str | None
    message_timeout: int | None
    min_job_age: int | None
    mpi_default: str | None
    mpi_params: Annotated[
        dict[str, bool | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    over_time_limit: int | str | None
    plugin_dir: Annotated[list[str] | None, Metadata(callback=ColonSepCallback)]
    plug_stack_config: str | None
    preempt_mode: str | None
    preempt_parameters: Annotated[
        dict[str, bool | int] | None,
        Metadata(callback=CommaDictCallback),
    ]
    preempt_type: str | None
    preempt_exempt_time: int | str | None
    pr_ep_parameters: Annotated[
        dict[str, Any] | None,
        Metadata(callback=CommaDictCallback),
    ]
    pr_ep_plugins: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    priority_calc_period: int
    priority_decay_half_life: int | str | None
    priority_favor_small: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    priority_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    priority_max_age: int | str | None
    priority_parameters: str | None
    priority_site_factor_parameters: str | None
    priority_site_factor_plugin: str | None
    priority_type: str | None
    priority_usage_reset_period: str | None
    priority_weight_age: int | None
    priority_weight_assoc: int | None
    priority_weight_fairshare: int | None
    priority_weight_job_size: int | None
    priority_weight_partition: int | None
    priority_weight_qos: int | None
    priority_weight_tres: Annotated[dict[str, int] | None, Metadata(callback=CommaDictCallback)]
    private_data: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    proctrack_type: str | None
    prolog: str | None
    prolog_epilog_timeout: int | None
    prolog_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    prolog_slurmctld: str | None
    propagate_prio_process: int | None
    propagate_resource_limits: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    propagate_resource_limits_except: Annotated[
        list[str] | None,
        Metadata(callback=CommaSepCallback),
    ]
    reboot_program: Annotated[str | None, Metadata(callback=QuoteCallback)]
    reconfig_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    requeue_exit: Annotated[list[str | int] | None, Metadata(callback=CommaSepCallback)]
    requeue_exit_hold: Annotated[list[str | int] | None, Metadata(callback=CommaSepCallback)]
    resume_fail_program: str | None
    resume_program: str | None
    resume_rate: int | None
    resume_timeout: int | None
    resv_epilog: str | None
    resv_over_run: int | str | None
    resv_prolog: str | None
    return_to_service: int | None
    scheduler_parameters: Annotated[
        dict[str, bool | int | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    scheduler_time_slice: int | None
    scheduler_type: str | None
    scron_parameters: Annotated[dict[str, bool] | None, Metadata(callback=CommaDictCallback)]
    select_type: str | None
    select_type_parameters: Annotated[
        dict[str, bool] | None,
        Metadata(callback=CommaDictCallback),
    ]
    slurmctld_addr: str | None
    slurmctld_debug: str | None
    slurmctld_host: Annotated[list[str] | None, Metadata(unique=False, callback=MultilineCallback)]
    slurmctld_log_file: str | None
    slurmctld_parameters: Annotated[
        dict[str, bool | int | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    slurmctld_pid_file: str | None
    slurmctld_port: int | str | None
    slurmctld_primary_off_prog: str | None
    slurmctld_primary_on_prog: str | None
    slurmctld_syslog_debug: str | None
    slurmctld_timeout: int | None
    slurmd_debug: str | None
    slurmd_log_file: str | None
    slurmd_parameters: Annotated[
        dict[str, bool | int] | None,
        Metadata(callback=CommaDictCallback),
    ]
    slurmd_pid_file: str | None
    slurmd_port: int | None
    slurmd_spool_dir: str | None
    slurmd_syslog_debug: str | None
    slurmd_timeout: int | None
    slurmd_user: str | None
    slurm_sched_log_file: str | None
    slurm_sched_log_level: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    slurm_user: str | None
    srun_epilog: str | None
    srun_port_range: str | None
    srun_prolog: str | None
    state_save_location: str | None
    suspend_exc_nodes: Annotated[
        dict[str, bool | int] | None,
        Metadata(callback=CommaDictColonPairCallback),
    ]
    suspend_exc_parts: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    suspend_exc_states: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    suspend_program: str | None
    suspend_rate: int | None
    suspend_time: int | str | None
    suspend_timeout: int | None
    switch_parameters: Annotated[
        dict[str, bool | int | str | list[str]] | None,
        Metadata(callback=CommaDictColonArrayCallback),
    ]
    switch_type: str | None
    task_epilog: str | None
    task_plugin: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    task_plugin_param: Annotated[
        dict[str, bool | str] | None,
        Metadata(callback=CommaDictCallback),
    ]
    task_prolog: str | None
    tcp_timeout: int | None
    tmpfs: str | None
    topology_param: Annotated[dict[str, bool | int] | None, Metadata(callback=CommaDictCallback)]
    topology_plugin: str | None
    track_wc_key: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    tree_width: int | None
    unkillable_step_program: str | None
    unkillable_step_timeout: int | None
    use_pam: Annotated[bool | None, Metadata(callback=IntBoolCallback)]
    v_size_factor: int | None
    wait_time: int | None
    x11_parameters: Annotated[dict[str, bool] | None, Metadata(callback=CommaDictCallback)]

    down_nodes: Annotated[
        DownNodesList, Metadata(unique=False, default_factory=lambda: DownNodesList())
    ]
    frontend_nodes: Annotated[
        FrontendNodeMapping,
        Metadata(
            origin="frontendnodes",
            alias="frontendnodename",
            unique=False,
            default_factory=lambda: FrontendNodeMapping(),
        ),
    ]
    nodes: Annotated[
        NodeMapping,
        Metadata(
            origin="nodes", alias="nodename", unique=False, default_factory=lambda: NodeMapping()
        ),
    ]
    nodesets: Annotated[
        NodeSetMapping,
        Metadata(
            origin="nodesets",
            alias="nodeset",
            unique=False,
            default_factory=lambda: NodeSetMapping(),
        ),
    ]
    partitions: Annotated[
        PartitionMapping,
        Metadata(
            origin="partitions",
            alias="partitionname",
            unique=False,
            default_factory=lambda: PartitionMapping(),
        ),
    ]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return SLURM_CONFIG_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]:  # noqa N805
        return _slurm_config_model_builder


class SlurmConfigEditor(BaseEditor):
    """Editor for the `slurm.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return SlurmConfig


_down_nodes_list_model_builder = make_model_builder(DownNodes)
_frontend_node_mapping_model_builder = make_model_builder(FrontendNode)
_node_mapping_model_builder = make_model_builder(Node)
_nodeset_mapping_model_builder = make_model_builder(NodeSet)
_partition_mapping_model_builder = make_model_builder(Partition)
_slurm_config_model_builder = make_model_builder(
    DownNodesList,
    FrontendNodeMapping,
    NodeMapping,
    NodeSetMapping,
    PartitionMapping,
)



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model representing the `acct_gather.conf` configuration file."""

__all__ = ["AcctGatherConfig", "AcctGatherConfigEditor"]

from typing import Annotated, Any

from .core.base import Metadata, Model, classproperty
from .core.callback import CommaSepCallback, SemicolonDictCallback, StrBoolCallback
from .core.editor import BaseEditor
from .core.schema import ACCT_GATHER_CONFIG_MODEL_SCHEMA


class AcctGatherConfig(Model):
    """Model representing the `acct_gather.conf` configuration file."""

    energy_ipmi_frequency: int | None
    energy_ipmi_calc_adjustment: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    energy_ipmi_power_sensors: Annotated[
        dict[str, list[int]] | None,
        Metadata(callback=SemicolonDictCallback),
    ]
    energy_ipmi_username: str | None
    energy_ipmi_password: str | None
    energy_ipmi_frequency: int | None
    energy_ipmi_timeout: int | None
    profile_hdf5_dir: str | None
    profile_hdf5_default: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    profile_influxdb_database: str | None
    profile_influxdb_default: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    profile_influxdb_host: str | None
    profile_influxdb_pass: str | None
    profile_influxdb_rt_policy: str | None
    profile_influxdb_user: str | None
    profile_influxdb_timeout: int | None
    infiniband_ofed_port: int | None
    sysfs_interfaces: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return ACCT_GATHER_CONFIG_MODEL_SCHEMA


class AcctGatherConfigEditor(BaseEditor):
    """Editor for the `acct_gather.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return AcctGatherConfig



#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the model and editor of the `slurmdbd.conf` configuration file."""

from constants import EXAMPLE_SLURMDBD_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import ModelError, SlurmdbdConfig, slurmdbdconfig

EXPECTED_SLURMDBD_DUMPS_OUTPUT = """
archiveevents=yes
archivejobs=yes
archiveresvs=yes
archivesteps=no
archivetxn=no
archiveusage=no
archivescript=/usr/sbin/slurm.dbd.archive
authinfo=socket=/var/run/munge/munge.socket.2
authtype=auth/munge
authalttypes=auth/jwt
authaltparameters=jwt_key=16549684561684@
dbdhost=slurmdbd-0
dbdbackuphost=slurmdbd-1
debuglevel=info
plugindir=/all/these/cool/plugins
purgeeventafter=1month
purgejobafter=12month
purgeresvafter=1month
purgestepafter=1month
purgesuspendafter=1month
purgetxnafter=12month
purgeusageafter=24month
logfile=/var/log/slurmdbd.log
pidfile=/var/run/slurmdbd.pid
slurmuser=slurm
storagepass=supersecretpasswd
storagetype=accounting_storage/mysql
storageuser=slurm
storagehost=127.0.0.1
storageport=3306
storageloc=slurm_acct_db
""".strip()


class TestSlurmdbdConfigEditor(TestCase):
    """Unit tests for the model and editor of the `slurmdbd.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/slurmdbd.conf", contents=EXAMPLE_SLURMDBD_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of the `slurmdbd.conf` model editor."""
        config = slurmdbdconfig.loads(EXAMPLE_SLURMDBD_CONFIG)
        self.assertListEqual(config.plugin_dir, ["/all/these/cool/plugins"])
        self.assertDictEqual(config.auth_alt_parameters, {"jwt_key": "16549684561684@"})
        self.assertEqual(config.slurm_user, "slurm")
        self.assertEqual(config.log_file, "/var/log/slurmdbd.log")

    def test_dumps(self) -> None:
        """Test `dumps` method of the `slurmdbd.conf` model editor."""
        config = slurmdbdconfig.loads(EXAMPLE_SLURMDBD_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(slurmdbdconfig.dumps(config), EXPECTED_SLURMDBD_DUMPS_OUTPUT)
        # New config and old config should not be equal since the editor strips all comments.
        self.assertNotEqual(slurmdbdconfig.dumps(config), EXAMPLE_SLURMDBD_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` context manager of the `slurmdbd.conf` model editor."""
        with slurmdbdconfig.edit("/etc/slurm/slurmdbd.conf") as config:
            config.archive_usage = True
            config.log_file = "/var/spool/slurmdbd.log"
            config.debug_flags = ["db_event", "db_job", "db_usage"]
            del config.auth_alt_types
            del config.auth_alt_parameters

        config = slurmdbdconfig.load("/etc/slurm/slurmdbd.conf")
        self.assertTrue(config.archive_usage)
        self.assertEqual(config.log_file, "/var/spool/slurmdbd.log")
        self.assertListEqual(config.debug_flags, ["db_event", "db_job", "db_usage"])
        self.assertIsNone(config.auth_alt_types)
        self.assertIsNone(config.auth_alt_parameters)

    def test_load_fail(self) -> None:
        """Test that bogus values are automatically caught when loading a new model."""
        # Catch if field value is unexpected. e.g. not a boolean value.
        with self.assertRaises(ValueError):
            SlurmdbdConfig.from_str(
                """
                archivesteps=totally
                """
            )

        # Catch if field is unexpected or in the incorrect syntax.
        with self.assertRaises(ModelError):
            SlurmdbdConfig.from_str(
                """
                archive_steps=yes
                """
            )

        # Catch if field value is invalid when checking against the model schema.
        with self.assertRaises(ModelError):
            SlurmdbdConfig.from_str(
                """
                LogFile=67
                """
            )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = slurmdbdconfig.loads(EXAMPLE_SLURMDBD_CONFIG)
            config.archive = True

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = slurmdbdconfig.loads(EXAMPLE_SLURMDBD_CONFIG)
            _ = config.archive



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from .core.base import Model, classproperty
from .core.editor import BaseEditor

class SlurmdbdConfig(Model):
    allow_no_def_acct: bool | None
    all_resources_absolute: bool | None
    archive_dir: str | None
    archive_events: bool | None
    archive_jobs: bool | None
    archive_resvs: bool | None
    archive_script: str | None
    archive_steps: str | None
    archive_suspend: str | None
    archive_txn: str | None
    archive_usage: str | None
    auth_alt_types: list[str] | None
    auth_alt_parameters: dict[str, str] | None
    auth_info: dict[str, str | bool] | None
    auth_type: str | None
    commit_delay: int | None
    communication_parameters: dict[str, int | bool] | None
    dbd_addr: str | None
    dbd_backup_host: str | None
    dbd_host: str | None
    dbd_port: int | None
    debug_flags: list[str] | None
    debug_level: str | None
    debug_level_syslog: str | None
    default_qos: str | None
    disable_coord_dbd: bool | None
    hash_plugin: str | None
    log_file: str | None
    log_time_format: str | None
    max_query_time_range: str | int | None
    message_timeout: str | None
    parameters: dict[str, bool] | None
    pid_file: str | None
    plugin_dir: list[str] | None
    private_data: list[str] | None
    purge_event_after: str | None
    purge_job_after: str | None
    purge_resv_after: str | None
    purge_step_after: str | None
    purge_suspend_after: str | None
    purge_txn_after: str | None
    purge_usage_after: str | None
    slurm_user: str | None
    storage_backup_host: str | None
    storage_host: str | None
    storage_loc: str | None
    storage_parameters: dict[str, str | list[str]] | None
    storage_pass: str | None
    storage_port: int | None
    storage_type: str | None
    storage_user: str | None
    tcp_timeout: int | None
    track_slurmctld_down: bool | None
    track_wc_key: bool | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804

class SlurmdbdConfigEditor(BaseEditor):
    @property
    def __model__(self) -> type[SlurmdbdConfig]: ...
    def dump(
        self,
        obj: SlurmdbdConfig,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None: ...
    def dumps(self, obj: SlurmdbdConfig, /) -> str: ...
    def load(self, file: str | os.PathLike, /) -> SlurmdbdConfig: ...
    def loads(self, s: str, /) -> SlurmdbdConfig: ...
    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[SlurmdbdConfig]: ...



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""JSON schemas for validating Slurm configuration models."""

__all__ = [
    "ACCT_GATHER_CONFIG_MODEL_SCHEMA",
    "CGROUP_CONFIG_MODEL_SCHEMA",
    "DOWN_NODES_MODEL_SCHEMA",
    "DOWN_NODES_LIST_MODEL_SCHEMA",
    "FRONTEND_NODE_MODEL_SCHEMA",
    "FRONTEND_NODE_MAPPING_MODEL_SCHEMA",
    "GRES_MODEL_SCHEMA",
    "GRES_LIST_MODEL_SCHEMA",
    "GRES_MAPPING_MODEL_SCHEMA",
    "GRES_CONFIG_MODEL_SCHEMA",
    "NODE_MODEL_SCHEMA",
    "NODE_MAPPING_MODEL_SCHEMA",
    "NODESET_MODEL_SCHEMA",
    "NODESET_MAPPING_MODEL_SCHEMA",
    "OCI_CONFIG_MODEL_SCHEMA",
    "PARTITION_MODEL_SCHEMA",
    "PARTITION_MAPPING_MODEL_SCHEMA",
    "SLURM_CONFIG_MODEL_SCHEMA",
    "SLURMDBD_CONFIG_MODEL_SCHEMA",
]

_GLOBAL_SCHEMA_VERSION = "https://json-schema.org/draft/2020-12/schema"

# `acct_gather.conf` data model schema.

ACCT_GATHER_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "energyipmicalcadjustment": {"type": "boolean"},
        "energyipmipowersensors": {
            "type": "object",
            "patternProperties": {
                "^.*$": {
                    "type": "array",
                    "items": {"type": "integer"},
                    "uniqueItems": True,
                }
            },
        },
        "energyipmiusername": {"type": "string"},
        "energyipmipassword": {"type": "string"},
        "energyipmifrequency": {"type": "integer", "minimum": 0},
        "energyipmitimeout": {"type": "integer", "minimum": 0},
        "profilehdf5dir": {"type": "string"},
        "profilehdf5default": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all",
                    "none",
                    "energy",
                    "filesystem",
                    "network",
                    "task",
                ],
            },
            "uniqueItems": True,
        },
        "profileinfluxdbdatabase": {"type": "string"},
        "profileinfluxdbdefault": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all",
                    "none",
                    "energy",
                    "filesystem",
                    "network",
                    "task",
                ],
            },
            "uniqueItems": True,
        },
        "profileinfluxdbhost": {"type": "string"},
        "profileinfluxdbpass": {"type": "string"},
        "profileinfluxdbrtpolicy": {"type": "string"},
        "profileinfluxdbuser": {"type": "string"},
        "profileinfluxdbtimeout": {"type": "integer", "minimum": 0},
        "infinibandofedport": {"type": "integer", "minimum": 0},
        "sysfsinterfaces": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
    },
    "additionalProperties": False,
}

# `cgroup.conf` data model schema.

CGROUP_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "cgroupmountpoint": {"type": "string"},
        "cgroupplugin": {
            "type": "string",
            "enum": ["cgroup/v1", "cgroup/v2", "autodetect", "disabled"],
        },
        "systemdtimeout": {"type": "integer", "minimum": 0},
        "ignoresystemd": {"type": "boolean"},
        "ignoreystemdonfailure": {"type": "boolean"},
        "enablecontrollers": {"type": "boolean"},
        "allowedramspace": {"type": "integer", "minimum": 0},
        "allowedswapspace": {"type": "integer", "minimum": 0},
        "constraincores": {"type": "boolean"},
        "constraindevices": {"type": "boolean"},
        "constrainramspace": {"type": "boolean"},
        "constrainswapspace": {"type": "boolean"},
        "maxrampercent": {"type": "number", "minimum": 0, "maximum": 100},
        "maxswappercent": {"type": "number", "minimum": 0, "maximum": 100},
        "memoryswappiness": {"type": "integer", "minimum": 0},
        "minramspace": {"type": "integer", "minimum": 0},
        "signalchildrenprocesses": {"type": "boolean"},
    },
    "additionalProperties": False,
}

# `gres.conf` data model schemas.

GRES_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "autodetect": {
            "type": "string",
            "enum": [
                "nrt",
                "nvidia",
                "nvml",
                "off",
                "oneapi",
                "rsmi",
            ],
        },
        "count": {
            "oneOf": [
                {"type": "string"},
                {"type": "integer", "minimum": 0},
            ],
        },
        "cores": {
            "type": "array",
            "items": {
                "oneOf": [
                    {"type": "string"},
                    {"type": "integer"},
                ],
            },
            "uniqueItems": True,
        },
        "file": {"type": "string"},
        "flags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "countonly",
                    "explicit",
                    "one_sharing",
                    "all_sharing",
                    "nvidia_gpu_env",
                    "amd_gpu_env",
                    "intel_gpu_env",
                    "opencl_env",
                    "no_gpu_env",
                ],
            },
            "uniqueItems": True,
        },
        "links": {
            "type": "array",
            "items": {
                "type": "integer",
                "minimum": -1,
            },
        },
        "multiplefiles": {
            "type": "array",
            "items": {
                "type": "string",
            },
            "uniqueItems": True,
        },
        "nodename": {
            "type": "string",
        },
        "type": {"type": "string"},
    },
    "additionalProperties": False,
}

GRES_LIST_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "array",
    "items": GRES_MODEL_SCHEMA,
    "uniqueItems": True,
}

GRES_MAPPING_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "patternProperties": {"^.+$": GRES_LIST_MODEL_SCHEMA},
}

GRES_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "autodetect": GRES_MODEL_SCHEMA["properties"]["autodetect"],
        "gres": GRES_MAPPING_MODEL_SCHEMA,
    },
    "additionalProperties": False,
}

# `oci.conf` data model schemas.

OCI_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "containerpath": {"type": "string"},
        "createenvfile": {"type": "string", "enum": ["null", "newline", "disable"]},
        "debugflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "accrue",
                    "agent",
                    "auditrpcs",
                    "backfill",
                    "backfillmap",
                    "burstbuffer",
                    "cgroup",
                    "conmgr",
                    "cpu_bind",
                    "cpufrequency",
                    "data",
                    "dbd_agent",
                    "dependency",
                    "energy",
                    "federation",
                    "frontend",
                    "gres",
                    "hetjob",
                    "gang",
                    "glob_silence",
                    "jobaccountgather",
                    "jobcomp",
                    "jobcontainer",
                    "license",
                    "network",
                    "networkraw",
                    "nodefeatures",
                    "no_conf_hash",
                    "power",
                    "priority",
                    "profile",
                    "protocol",
                    "reservation",
                    "route",
                    "script",
                    "selecttype",
                    "steps",
                    "switch",
                    "tls",
                    "tracejobs",
                    "triggers",
                ],
            },
            "uniqueItems": True,
        },
        "disablecleanup": {"type": "boolean"},
        "disablehooks": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "envexclude": {"type": "string"},
        "mountspooldir": {"type": "string"},
        "runtimeenvexclude": {"type": "string"},
        "filedebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "ignorefileconfigjson": {"type": "boolean"},
        "runtimecreate": {"type": "string"},
        "runtimedelete": {"type": "string"},
        "runtimekill": {"type": "string"},
        "runtimequery": {"type": "string"},
        "runtimerun": {"type": "string"},
        "runtimestart": {"type": "string"},
        "srunpath": {"type": "string"},
        "srunargs": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "stdiodebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "syslogdebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
    },
    "additionalProperties": False,
}

# `slurm.conf` data model schemas.

DOWN_NODES_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "downnodes": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "reason": {"type": "string"},
        "state": {"type": "string", "enum": ["down", "drain", "fail", "failing", "future"]},
    },
    "additionalProperties": False,
}

FRONTEND_NODE_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "frontendname": {"type": "string"},
        "frontendaddr": {"type": "string"},
        "allowgroups": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "allowusers": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "denygroups": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "denyusers": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "port": {"type": "integer", "minimum": 0},
        "reason": {"type": "string"},
        "state": {"type": "string", "enum": ["down", "drain", "fail", "failing", "unknown"]},
    },
    "additionalProperties": False,
}

NODE_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "nodename": {"type": "string"},
        "nodehostname": {"type": "string"},
        "nodeaddr": {"type": "string"},
        "bcastaddr": {"type": "string"},
        "boards": {"type": "integer", "minimum": 1},
        "corespeccount": {"type": "integer", "minimum": 1},
        "corespersocket": {"type": "integer", "minimum": 1},
        "cpubind": {"type": "string", "enum": ["none", "socket", "ldom", "core", "thread"]},
        "cpus": {"type": "integer", "minimum": 0},
        "cpuspeclist": {"type": "array", "items": {"type": "integer"}, "uniqueItems": True},
        "features": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "gres": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "memspeclimit": {"type": "integer", "minimum": 1},
        "port": {"type": "integer"},
        "procs": {"type": "integer", "minimum": 1},
        "realmemory": {"type": "integer", "minimum": 1},
        "reason": {"type": "string"},
        "sockets": {"type": "integer", "minimum": 1},
        "socketsperboard": {"type": "integer", "minimum": 1},
        "state": {
            "type": "string",
            "enum": ["cloud", "down", "drain", "fail", "failing", "future", "unknown"],
        },
        "threadspercore": {"type": "integer", "minimum": 1},
        "tmpdisk": {"type": "integer", "minimum": 1},
        "weight": {"type": "integer", "minimum": 1},
    },
    "additionalProperties": False,
}

NODESET_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "nodeset": {"type": "string"},
        "feature": {"type": "string"},
        "nodes": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
    },
    "additionalProperties": False,
}

PARTITION_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "partitionname": {"type": "string"},
        "allocnodes": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "allowaccounts": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "allowgroups": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "allowqos": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "alternate": {"type": "string"},
        "cpubind": {"type": "string", "enum": ["none", "socket", "ldom", "core", "thread"]},
        "default": {"type": "boolean"},
        "defaulttime": {
            "oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}],
        },
        "defcpupergpu": {"type": "integer", "minimum": 0},
        "defmempercpu": {"type": "integer", "minimum": 0},
        "defmempergpu": {"type": "integer", "minimum": 0},
        "defmempernode": {"type": "integer", "minimum": 0},
        "denyaccounts": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "denyqos": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "disablerootjobs": {"type": "boolean"},
        "exclusivetopo": {"type": "boolean"},
        "exclusiveuser": {"type": "boolean"},
        "gracetime": {"type": "number", "minimum": 0},
        "hidden": {"type": "boolean"},
        "lln": {"type": "boolean"},
        "maxcpuspernode": {"type": "integer", "minimum": 0},
        "maxcpuspersocket": {"type": "integer", "minimum": 0},
        "maxmempercpu": {"type": "integer", "minimum": 0},
        "maxmempernode": {"type": "integer", "minimum": 0},
        "maxnodes": {
            "oneOf": [
                {"type": "integer", "minimum": -1},
                {"type": "string", "enum": ["unlimited"]},
            ]
        },
        "maxtime": {
            "oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}],
        },
        "minnodes": {"type": "integer", "minimum": 0},
        "nodes": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "oversubscribe": {"type": "string", "enum": ["exclusive", "force", "yes", "no"]},
        "overtimelimit": {
            "oneOf": [
                {"type": "integer", "minimum": 0, "maximum": 65533},
                {"type": "string", "enum": ["unlimited"]},
            ]
        },
        "powerdownonidle": {"type": "boolean"},
        "preemptmode": {"type": "string", "enum": ["off", "cancel", "gang", "requeue", "suspend"]},
        "priorityjobfactor": {"type": "integer", "maximum": 65533},
        "prioritytier": {"type": "integer", "maximum": 65533},
        "qos": {"type": "string"},
        "reqresv": {"type": "boolean"},
        "resumetimeout": {"type": "integer", "minimum": 0},
        "rootonly": {"type": "boolean"},
        "selecttypeparameters": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["cr_core", "cr_core_memory", "cr_socket", "cr_socket_memory"],
            },
        },
        "state": {"type": "string", "enum": ["up", "down", "drain", "inactive"]},
        "suspendtime": {
            "oneOf": [
                {"type": "integer", "minimum": -1},
                {"type": "string", "enum": ["infinite"]},
            ],
        },
        "suspendtimeout": {"type": "integer", "minimum": 0},
        "tresbillingweights": {
            "type": "object",
            "patternProperties": {"^.*$": {"type": "number"}},
        },
    },
    "additionalProperties": False,
}

NODE_MAPPING_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "patternProperties": {"^.+$": NODE_MODEL_SCHEMA},
}

DOWN_NODES_LIST_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "array",
    "items": DOWN_NODES_MODEL_SCHEMA,
    "uniqueItems": True,
}

FRONTEND_NODE_MAPPING_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "patternProperties": {"^.+$": FRONTEND_NODE_MODEL_SCHEMA},
}

NODESET_MAPPING_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "patternProperties": {"^.+$": NODESET_MODEL_SCHEMA},
}

PARTITION_MAPPING_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "patternProperties": {"^.+$": PARTITION_MODEL_SCHEMA},
}

SLURM_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "accountingstoragebackuphost": {"type": "string"},
        "accountingstorageenforce": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all",
                    "associations",
                    "limits",
                    "nojobs",
                    "nosteps",
                    "qos",
                    "safe",
                    "wckeys",
                ],
            },
            "uniqueItems": True,
        },
        "accountingstorageexternalhost": {"type": "string"},
        "accountingstoragehost": {"type": "string"},
        "accountingstorageparameters": {
            "type": "object",
            "properties": {
                "max_step_records": {"type": "integer"},
                "ssl_cert": {"type": "string"},
                "ssl_ca": {"type": "string"},
                "ssl_capath": {"type": "string"},
                "ssl_key": {"type": "string"},
                "ssl_cipher": {
                    "type": "string",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                },
            },
            "additionalProperties": False,
        },
        "accountingstoragepass": {"type": "string"},
        "accountingstorageport": {"type": "integer", "minimum": 0},
        "accountingstoragetres": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "accountingstoragetype": {"type": "string", "enum": ["accounting_storage/slurmdbd"]},
        "accountingstorageuser": {"type": "string"},
        "accountingstoreflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["job_comment", "job_env", "job_extra", "job_script", "no_stdio"],
            },
        },
        "acctgathernodefreq": {"type": "integer", "minimum": 0},
        "acctgatherenergytype": {
            "type": "string",
            "enum": [
                "acct_gather_energy/gpu",
                "acct_gather_energy/ipmi",
                "acct_gather_energy/pm_counters",
                "acct_gather_energy/rapl",
                "acct_gather_energy/xcc",
            ],
        },
        "acctgatherinterconnecttype": {
            "type": "string",
            "enum": ["acct_gather_interconnect/ofed", "acct_gather_interconnect/sysfs"],
        },
        "acctgatherfilesystemtype": {"type": "string", "enum": ["acct_gather_filesystem/lustre"]},
        "acctgatherprofiletype": {
            "type": "string",
            "enum": ["acct_gather_profile/hdf5", "acct_gather_profile/influxdb"],
        },
        "allowspecresourcesusage": {"type": "boolean"},
        "authalttypes": {"type": "array", "items": {"type": "string", "enum": ["auth/jwt"]}},
        "authaltparameters": {
            "type": "object",
            "properties": {
                "disable_token_creation": {"type": "boolean"},
                "max_token_lifespan": {"type": "integer"},
                "jwks": {"type": "string"},
                "jwt_key": {"type": "string"},
                "userclaimfield": {"type": "string"},
            },
            "additionalProperties": False,
        },
        "authinfo": {
            "type": "object",
            "properties": {
                "cred_expire": {"type": "integer"},
                "socket": {"type": "string"},
                "ttl": {"type": "integer"},
                "use_client_ids": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "authtype": {"type": "string", "enum": ["auth/munge", "auth/slurm"]},
        "batchstarttimeout": {"type": "integer", "minimum": 0},
        "bcastexclude": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "bcastparameters": {
            "type": "object",
            "properties": {
                "destdir": {"type": "string"},
                "compression": {"type": "string", "enum": ["lz4", "none"]},
                "send_libs": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "burstbuffertype": {
            "type": "string",
            "enum": ["burst_buffer/datawarp", "burst_buffer/lua", "burst_buffer/none"],
        },
        "clifilterplugins": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["cli_filter/lua", "cli_filter/syslog", "cli_filter/user_defaults"],
            },
            "uniqueItems": True,
        },
        "clustername": {"type": "string"},
        "communicationparameters": {
            "type": "object",
            "properties": {
                "block_null_hash": {"type": "boolean"},
                "disableipv4": {"type": "boolean"},
                "enableipv6": {"type": "boolean"},
                "getnameinfo_cache_timeout": {"type": "integer", "minimum": 0},
                "keepaliveinterval": {"type": "integer", "minimum": 0},
                "keepaliveprobes": {"type": "integer", "minimum": 0},
                "keepalivetime": {"type": "integer", "minimum": 0},
                "noctldinaddrany": {"type": "boolean"},
                "noinaddrany": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "completewait": {"type": "integer", "minimum": 0, "maximum": 65533},
        "cpufreqdef": {
            "type": "string",
            "enum": ["conservative", "ondemand", "performance", "powersave"],
        },
        "cpufreqgovernors": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "conservative",
                    "ondemand",
                    "performance",
                    "powersave",
                    "schedutil",
                    "userspace",
                ],
            },
            "uniqueItems": True,
        },
        "credtype": {"type": "string", "enum": ["cred/munge", "cred/slurm"]},
        "debugflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "accrue",
                    "agent",
                    "auditrpcs",
                    "backfill",
                    "backfillmap",
                    "burstbuffer",
                    "cgroup",
                    "conmgr",
                    "cpu_bind",
                    "cpufrequency",
                    "data",
                    "dbd_agent",
                    "dependency",
                    "energy",
                    "federation",
                    "frontend",
                    "gres",
                    "hetjob",
                    "gang",
                    "glob_silence",
                    "jobaccountgather",
                    "jobcomp",
                    "jobcontainer",
                    "license",
                    "network",
                    "networkraw",
                    "nodefeatures",
                    "no_conf_hash",
                    "power",
                    "priority",
                    "profile",
                    "protocol",
                    "reservation",
                    "route",
                    "script",
                    "selecttype",
                    "steps",
                    "switch",
                    "tls",
                    "tracejobs",
                    "triggers",
                ],
            },
            "uniqueItems": True,
        },
        "defcpupergpu": {"type": "integer", "minimum": 0},
        "defmempercpu": {"type": "integer", "minimum": 0},
        "defmempergpu": {"type": "integer", "minimum": 0},
        "defmempernode": {"type": "integer", "minimum": 0},
        "dependencyparameters": {
            "type": "object",
            "properties": {
                "disable_remote_singleton": {"type": "boolean"},
                "kill_invalid_depend": {"type": "boolean"},
                "max_depend_depth": {"type": "integer", "minimum": 0},
            },
            "additionalProperties": False,
        },
        "disablerootjobs": {"type": "boolean"},
        "eiotimeout": {"type": "integer", "minimum": 0, "maximum": 65533},
        "enforcepartlimits": {"type": "string", "enum": ["all", "any", "no"]},
        "epilog": {"type": "string"},
        "epilogmsgtime": {"type": "integer", "minimum": 0},
        "epilogslurmctld": {"type": "string"},
        "fairsharedampeningfactor": {"type": "integer", "minimum": 1},
        "federationparameters": {
            "type": "object",
            "properties": {
                "fed_display": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "firstjobid": {"type": "integer", "minimum": 1},
        "getenvtimeout": {"type": "integer", "minimum": 0},
        "grestypes": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "groupupdateforce": {"type": "boolean"},
        "groupupdatetime": {"type": "integer", "minimum": 0},
        "gpufreqdef": {
            "type": "array",
            "items": {
                "oneOf": [
                    {"type": "string"},
                    {"type": "integer"},
                ]
            },
            "uniqueItems": True,
        },
        "hashplugin": {"type": "string", "enum": ["hash/k12", "hash/sha3"]},
        "healthcheckinterval": {"type": "integer", "minimum": 0},
        "healthchecknodestate": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["alloc", "any", "cycle", "idle", "nondrained_idle", "mixed"],
            },
            "uniqueItems": True,
        },
        "healthcheckprogram": {"type": "string"},
        "inactivelimit": {"type": "integer", "minimum": 0, "maximum": 65533},
        "include": {
            "type": "array",
            "items": {
                "type": "string",
            },
            "uniqueItems": True,
        },
        "interactivestepoptions": {"type": "string"},
        "jobacctgathertype": {
            "type": "string",
            "enum": ["jobacct_gather/cgroup", "jobacct_gather/linux", "jobacct_gather/none"],
        },
        "jobacctgatherfrequency": {
            "type": "object",
            "properties": {
                "task": {"type": "integer", "minimum": 0},
                "energy": {"type": "integer", "minimum": 0},
                "network": {"type": "integer", "minimum": 0},
                "filesystem": {"type": "integer", "minimum": 0},
            },
            "additionalProperties": False,
        },
        "jobacctgatherparams": {
            "type": "object",
            "properties": {
                "noshared": {"type": "boolean"},
                "usepss": {"type": "boolean"},
                "overmemorykill": {"type": "boolean"},
                "disablegpuacct": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "jobcomphost": {"type": "string"},
        "jobcomploc": {"type": "string"},
        "jobcompparams": {
            "type": "object",
            "properties": {
                "flush_timeout": {"type": "integer", "minimum": 0},
                "poll_interval": {"type": "integer", "minimum": 0},
                "requeue_on_msg_timeout": {"type": "boolean"},
                "topic": {"type": "string"},
            },
            "additionalProperties": False,
        },
        "jobcomppass": {"type": "string"},
        "jobcompport": {"type": "integer", "minimum": 0},
        "jobcomptype": {
            "type": "string",
            "enum": [
                "jobcomp/none",
                "jobcomp/elasticsearch",
                "jobcomp/filetxt",
                "jobcomp/kafka",
                "jobcomp/lua",
                "jobcomp/mysql",
                "jobcomp/script",
            ],
        },
        "jobcompuser": {"type": "string"},
        "jobcontainertype": {"type": "string", "enum": ["job_container/tmpfs"]},
        "jobfileappend": {"type": "boolean"},
        "jobrequeue": {"type": "boolean"},
        "jobsubmitplugins": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all_partitions",
                    "defaults",
                    "logging",
                    "lua",
                    "partition",
                    "pbs",
                    "require_timelimit",
                ],
            },
            "uniqueItems": True,
        },
        "killonbadexit": {"type": "boolean"},
        "killwait": {"type": "integer", "minimum": 0, "maximum": 65533},
        "maxbatchrequeue": {"type": "integer", "minimum": 0},
        "nodefeaturesplugins": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "node_features/knl_cray",
                    "node_features/knl_generic",
                    "node_features/helpers",
                ],
            },
            "uniqueItems": True,
        },
        "launchparameters": {
            "type": "object",
            "properties": {
                "batch_step_set_cpu_freq": {"type": "boolean"},
                "cray_net_exclusive": {"type": "boolean"},
                "enable_nss_slurm": {"type": "boolean"},
                "lustre_no_flush": {"type": "boolean"},
                "mem_sort": {"type": "boolean"},
                "mpir_use_nodeaddr": {"type": "boolean"},
                "disable_send_gids": {"type": "boolean"},
                "slurmstepd_memlock_all": {"type": "boolean"},
                "test_exec": {"type": "boolean"},
                "use_interactive_step": {"type": "boolean"},
                "ulimit_pam_adopt": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "licenses": {
            "type": "object",
            "patternProperties": {
                "^.*$": {
                    "type": "integer",
                    "minimum": -1,
                },
            },
        },
        "logtimeformat": {
            "type": "string",
            "enum": [
                "iso8601",
                "iso8601,format_stderr",
                "iso8601_ms",
                "iso8601_ms,format_stderr",
                "rfc5424",
                "rfc5424,format_stderr",
                "rfc5424_ms",
                "rfc5424_ms,format_stderr",
                "rfc3339",
                "rfc3339,format_stderr",
                "clock",
                "clock,format_stderr",
                "short",
                "short,format_stderr",
                "thread_id",
                "thread_id,format_stderr",
            ],
        },
        "maildomain": {"type": "string"},
        "mailprog": {"type": "string"},
        "maxarraysize": {"type": "integer", "minimum": 0, "maximum": 4000001},
        "maxdbdmsgs": {"type": "integer", "minimum": 10000},
        "maxjobcount": {"type": "integer", "minimum": 0},
        "maxjobid": {"type": "integer", "minimum": 0},
        "maxmempercpu": {"type": "integer", "minimum": 0},
        "maxmempernode": {"type": "integer", "minimum": 0},
        "maxnodecount": {"type": "integer", "minimum": 0, "maximum": 65533},
        "maxstepcount": {"type": "integer", "minimum": 0},
        "maxtaskspernode": {"type": "integer", "minimum": 0, "maximum": 65533},
        "mcsparameters": {
            "type": "object",
            "properties": {
                "ondemand": {"type": "boolean"},
                "enforced": {"type": "boolean"},
                "noselect": {"type": "boolean"},
                "select": {"type": "boolean"},
                "ondemandselect": {"type": "boolean"},
                "privatedata": {"type": "boolean"},
                "mcs_plugin_parameters": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                },
            },
            "additionalProperties": False,
        },
        "mcsplugin": {
            "type": "string",
            "enum": ["mcs/none", "mcs/account", "mcs/group", "mcs/user", "mcs/label"],
        },
        "messagetimeout": {"type": "integer", "minimum": 0},
        "minjobage": {"type": "integer", "minimum": 0},
        "mpidefault": {"type": "string", "enum": ["pmi2", "pmix", "none"]},
        "mpiparams": {
            "type": "object",
            "properties": {
                "ports": {"type": "string"},
                "disable_slurm_hydra_bootstrap": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "overtimelimit": {
            "oneOf": [
                {"type": "integer", "minimum": 0, "maximum": 65533},
                {"type": "string", "enum": ["unlimited"]},
            ]
        },
        "plugindir": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "plugstackconfig": {"type": "string"},
        "preemptmode": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["off", "cancel", "gang", "requeue", "suspend", "within"],
            },
        },
        "preemptparameters": {
            "type": "object",
            "properties": {
                "min_exempt_priority": {"type": "integer"},
                "reclaim_licenses": {"type": "boolean"},
                "reorder_count": {"type": "integer", "minimum": 0},
                "send_user_signal": {"type": "boolean"},
                "strict_order": {"type": "boolean"},
                "suspend_grace_time": {"type": "boolean"},
                "youngest_first": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "preempttype": {
            "type": "string",
            "enum": ["preempt/partition_prio", "preempt/qos"],
        },
        "preemptexempttime": {
            "oneOf": [{"type": "integer", "minimum": -1}, {"type": "string"}],
        },
        "prepparameters": {"type": "string"},
        "prepplugins": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "prioritycalcperiod": {"type": "integer", "minimum": 0},
        "prioritydecayhalflife": {
            "oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]
        },
        "priorityfavorsmall": {"type": "boolean"},
        "priorityflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "accrue_always",
                    "calculate_running",
                    "depth_oblivious",
                    "no_fair_tree",
                    "incr_only",
                    "max_tres",
                    "no_normal_all",
                    "no_normal_assoc",
                    "no_normal_part",
                    "no_normal_qos",
                    "no_normal_tres",
                    "small_relative_to_time",
                ],
            },
            "uniqueItems": True,
        },
        "prioritymaxage": {"oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]},
        "priorityparameters": {"type": "string"},
        "prioritysitefactorparameters": {"type": "string"},
        "prioritysitefactorplugin": {"type": "string"},
        "prioritytype": {"type": "string", "enum": ["priority/basic", "priority/multifactor"]},
        "priorityusageresetperiod": {
            "type": "string",
            "enum": [
                "none",
                "now",
                "daily",
                "weekly",
                "monthly",
                "quarterly",
                "yearly",
            ],
        },
        "priorityweightage": {"type": "integer", "minimum": 0},
        "priorityweightassoc": {"type": "integer", "minimum": 0},
        "priorityweightfairshare": {"type": "integer", "minimum": 0},
        "priorityweightjobsize": {"type": "integer", "minimum": 0},
        "priorityweightpartition": {"type": "integer", "minimum": 0},
        "priorityweightqos": {"type": "integer", "minimum": 0},
        "priorityweighttres": {
            "type": "object",
            "patternProperties": {"^.*$": {"type": "integer"}},
        },
        "privatedata": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "accounts",
                    "events",
                    "jobs",
                    "nodes",
                    "partitions",
                    "reservations",
                    "usage",
                    "users",
                ],
            },
            "uniqueItems": True,
        },
        "proctracktype": {
            "type": "string",
            "enum": ["proctrack/cgroup", "proctrack/linuxproc", "proctrack/pgid"],
        },
        "prolog": {"type": "string"},
        "prologepilogtimeout": {"type": "integer", "minimum": 0},
        "prologflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "alloc",
                    "contain",
                    "deferbatch",
                    "nohold",
                    "forcerequeueonfail",
                    "runinjob",
                    "serial",
                    "x11",
                ],
            },
            "uniqueItems": True,
        },
        "prologslurmctld": {"type": "string"},
        "propagateprioprocess": {"type": "integer", "minimum": 0, "maximum": 2},
        "propagateresourcelimits": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all",
                    "none",
                    "as",
                    "core",
                    "cpu",
                    "data",
                    "fsize",
                    "memlock",
                    "nofile",
                    "nproc",
                    "rss",
                    "stack",
                ],
            },
            "uniqueItems": True,
        },
        "propagateresourcelimitsexcept": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "all",
                    "none",
                    "as",
                    "core",
                    "cpu",
                    "data",
                    "fsize",
                    "memlock",
                    "nofile",
                    "nproc",
                    "rss",
                    "stack",
                ],
            },
            "uniqueItems": True,
        },
        "rebootprogram": {"type": "string"},
        "reconfigflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["keeppartinfo", "keeppartstate", "keeppowersavesettings"],
            },
            "uniqueItems": True,
        },
        "requeueexit": {
            "type": "array",
            "items": {"oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]},
            "uniqueItems": True,
        },
        "requeueexithold": {
            "type": "array",
            "items": {"oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]},
            "uniqueItems": True,
        },
        "resumefailprogram": {"type": "string"},
        "resumeprogram": {"type": "string"},
        "resumerate": {"type": "integer", "minimum": 0},
        "resumetimeout": {"type": "integer", "minimum": 0},
        "resvepilog": {"type": "string"},
        "resvoverrun": {
            "oneOf": [
                {"type": "integer", "minimum": 0, "maximum": 65533},
                {"type": "string", "enum": ["unlimited"]},
            ]
        },
        "resvprolog": {"type": "string"},
        "returntoservice": {"type": "integer", "minimum": 0, "maximum": 2},
        "schedulerparameters": {
            "type": "boolean",
            "properties": {
                "allow_zero_lic": {"type": "boolean"},
                "assoc_limit_stop": {"type": "boolean"},
                "batch_sched_delay": {"type": "integer", "minimum": 0},
                "bb_array_stage_cnt": {"type": "integer", "minimum": 0},
                "bf_allow_magnetic_slot": {"type": "boolean"},
                "bf_busy_nodes": {"type": "boolean"},
                "bf_continue": {"type": "boolean"},
                "bf_hetjob_immediate": {"type": "boolean"},
                "bf_hetjob_prio": {"type": "string", "enum": ["min", "avg", "max"]},
                "bf_interval": {"type": "integer", "minimum": -1, "maximum": 10800},
                "bf_job_part_count_reserve": {"type": "integer", "minimum": 0, "maximum": 100000},
                "bf_licenses": {"type": "boolean"},
                "bf_max_job_array_resv": {"type": "integer", "minimum": 0, "maximum": 1000},
                "bf_max_job_assoc": {"type": "integer", "minimum": 0, "maximum": 1000000},
                "bf_max_job_part": {"type": "integer", "minimum": 0, "maximum": 1000000},
                "bf_max_job_start": {"type": "integer", "minimum": 0, "maximum": 10000},
                "bf_max_job_test": {"type": "integer", "minimum": 1, "maximum": 1000000},
                "bf_max_job_user": {"type": "integer", "minimum": 0},
                "bf_max_job_user_part": {"type": "integer", "minimum": 0},
                "bf_max_time": {"type": "integer", "minimum": 1, "maximum": 3600},
                "bf_min_age_reserve": {"type": "integer", "minimum": 2592000},
                "bf_min_prio_reserve": {"type": "integer", "minimum": 0, "maximum": 2**63},
                "bf_node_space_size": {"type": "integer", "minimum": 1, "maximum": 2000000},
                "bf_one_resv_per_job": {"type": "boolean"},
                "bf_resolution": {"type": "integer", "minimum": 1, "maximum": 3600},
                "bf_running_job_reserve": {"type": "boolean"},
                "bf_topopt_enable": {"type": "boolean"},
                "bf_topopt_iterations": {"type": "integer", "minimum": 0},
                "bf_window": {"type": "integer", "minimum": 1, "maximum": 43200},
                "bf_window_linear": {"type": "integer", "minimum": 0},
                "bf_yield_interval": {"type": "integer", "minimum": 1, "maximum": 10000000},
                "bf_yield_sleep": {"type": "integer", "minimum": 1, "maximum": 10000000},
                "build_queue_timeout": {"type": "integer", "minimum": 0},
                "correspond_after_task_cnt": {"type": "integer", "minimum": 0},
                "default_queue_depth": {"type": "integer", "minimum": 0},
                "defer": {"type": "boolean"},
                "defer_batch": {"type": "boolean"},
                "delay_boot": {"type": "integer", "minimum": 0},
                "disable_job_shrink": {"type": "boolean"},
                "disable_hetjob_steps": {"type": "boolean"},
                "enable_hetjob_steps": {"type": "boolean"},
                "enable_job_state_cache": {"type": "boolean"},
                "enable_user_top": {"type": "boolean"},
                "extra_constraints": {"type": "boolean"},
                "ignore_constraint_validation": {"type": "boolean"},
                "ignore_numa": {"type": "boolean"},
                "ignore_prefer_validation": {"type": "boolean"},
                "max_array_tasks": {"type": "integer", "minimum": 0},
                "max_rpc_cnt": {"type": "integer", "minimum": 0},
                "max_sched_time": {"type": "integer", "minimum": 0},
                "max_script_size": {"type": "integer", "minimum": 0},
                "max_submit_line_size": {"type": "integer", "minimum": 0, "maximum": 2},
                "max_switch_wait": {"type": "integer", "minimum": 0},
                "no_backup_scheduling": {"type": "boolean"},
                "no_env_cache": {"type": "boolean"},
                "nohold_on_prolog_fail": {"type": "boolean"},
                "pack_serial_at_end": {"type": "boolean"},
                "partition_job_depth": {"type": "integer"},
                "reduce_completing_frag": {"type": "boolean"},
                "requeue_setup_env_fail": {"type": "boolean"},
                "salloc_wait_nodes": {"type": "boolean"},
                "sbatch_wait_nodes": {"type": "boolean"},
                "sched_interval": {"type": "integer", "minimum": -1},
                "sched_max_job_start": {"type": "integer", "minimum": 0},
                "sched_min_interval": {"type": "integer", "minimum": 0},
                "spec_cores_first": {"type": "boolean"},
                "step_retry_count": {"type": "integer", "minimum": 0},
                "step_retry_time": {"type": "integer", "minimum": 0},
                "time_min_as_soft_limit": {"type": "boolean"},
                "whole_hetjob": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "schedulertimeslice": {"type": "integer", "minimum": 5, "maximum": 65533},
        "schedulertype": {
            "type": "string",
            "enum": ["sched/backfill", "sched/builtin"],
        },
        "scronparameters": {
            "type": "object",
            "properties": {
                "enable": {"type": "boolean"},
                "explicit_scancel": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "selecttype": {
            "type": "string",
            "enum": ["select/cons_tres", "select/linear"],
        },
        "selecttypeparameters": {
            "type": "object",
            "properties": {
                "cr_cpu": {"type": "boolean"},
                "cr_cpu_memory": {"type": "boolean"},
                "cr_core": {"type": "boolean"},
                "cr_core_memory": {"type": "boolean"},
                "cr_one_task_per_core": {"type": "boolean"},
                "cr_core_default_dist_block": {"type": "boolean"},
                "cr_lln": {"type": "boolean"},
                "cr_pack_nodes": {"type": "boolean"},
                "ll_shared_gres": {"type": "boolean"},
                "cr_socket": {"type": "boolean"},
                "cr_socket_memory": {"type": "boolean"},
                "cr_memory": {"type": "boolean"},
                "multiple_sharing_gres_pj": {"type": "boolean"},
                "enforce_binding_gres": {"type": "boolean"},
                "one_task_per_sharing_gres": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "slurmctldaddr": {"type": "string"},
        "slurmctlddebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "slurmctldhost": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "slurmctldlogfile": {"type": "string"},
        "slurmctldparameters": {
            "type": "object",
            "properties": {
                "allow_user_triggers": {"type": "boolean"},
                "cloud_dns": {"type": "boolean"},
                "conmgr_max_connections": {"type": "integer", "minimum": 0},
                "conmgr_threads": {"type": "integer", "minimum": 0},
                "conmgr_use_poll": {"type": "boolean"},
                "conmgr_connect_timeout": {"type": "integer", "minimum": 0},
                "conmgr_read_timeout": {"type": "integer", "minimum": 0},
                "conmgr_wait_write_delay": {"type": "integer", "minimum": 0},
                "conmgr_write_timeout": {"type": "integer", "minimum": 0},
                "disable_triggers": {"type": "boolean"},
                "enable_configless": {"type": "boolean"},
                "idle_on_node_suspend": {"type": "boolean"},
                "node_reg_mem_percent": {"type": "integer", "minimum": 0},
                "no_quick_restart": {"type": "boolean"},
                "power_save_interval": {"type": "integer", "minimum": 0},
                "power_save_min_interval": {"type": "integer", "minimum": 0},
                "max_powered_nodes": {"type": "integer", "minimum": 0},
                "max_dbd_msg_action": {"type": "string", "enum": ["discard", "exit"]},
                "reboot_from_controller": {"type": "boolean"},
                "rl_bucket_size": {"type": "integer", "minimum": 0},
                "rl_enable": {"type": "boolean"},
                "rl_log_freq": {"type": "integer", "minimum": -1},
                "rl_refill_period": {"type": "integer", "minimum": 0},
                "rl_refill_rate": {"type": "integer", "minimum": 0},
                "rl_table_size": {"type": "integer", "minimum": 0},
                "enable_stepmgr": {"type": "boolean"},
                "user_resv_delete": {"type": "boolean"},
                "validate_nodeaddr_threads": {"type": "integer", "minimum": 0, "maximum": 64},
            },
            "additionalProperties": False,
        },
        "slurmctldpidfile": {"type": "string"},
        "slurmctldport": {"oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]},
        "slurmctldprimaryoffprog": {"type": "string"},
        "slurmctldprimaryonprog": {"type": "string"},
        "slurmctldsyslogdebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "slurmctldtimeout": {"type": "integer", "minimum": 0, "maximum": 65533},
        "slurmddebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "slurmdlogfile": {"type": "string"},
        "slurmdparameters": {
            "type": "object",
            "properties": {
                "allow_ecores": {"type": "boolean"},
                "config_overrides": {"type": "boolean"},
                "conmgr_max_connections": {"type": "integer", "minimum": 0},
                "conmgr_threads": {"type": "integer", "minimum": 0},
                "conmgr_use_poll": {"type": "boolean"},
                "conmgr_connect_timeout": {"type": "integer", "minimum": 0},
                "conmgr_read_timeout": {"type": "integer", "minimum": 0},
                "conmgr_wait_write_delay": {"type": "integer", "minimum": 0},
                "conmgr_write_timeout": {"type": "integer", "minimum": 0},
                "l3cache_as_socket": {"type": "boolean"},
                "numa_node_as_socket": {"type": "boolean"},
                "shutdown_on_reboot": {"type": "boolean"},
                "contain_spank": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "slurmdpidfile": {"type": "string"},
        "slurmdport": {"type": "integer", "minimum": 0},
        "slurmdspooldir": {"type": "string"},
        "slurmdsyslogdebug": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "slurmdtimeout": {"type": "integer", "minimum": 0, "maximum": 65533},
        "slurmduser": {"type": "string"},
        "slurmschedlogfile": {"type": "string"},
        "slurmschedloglevel": {"type": "boolean"},
        "slurmuser": {"type": "string"},
        "srunepilog": {"type": "string"},
        "srunportrange": {"type": "string"},
        "srunprolog": {"type": "string"},
        "statesavelocation": {"type": "string"},
        "suspendexcnodes": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "suspendexcparts": {
            "type": "array",
            "items": {"type": "string"},
            "uniqueItems": True,
        },
        "suspendexcstates": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "cloud",
                    "down",
                    "drain",
                    "dynamic_future",
                    "dynamic_norm",
                    "fail",
                    "invalid_reg",
                    "maintenance",
                    "not_responding",
                    "perfctrs",
                    "planned",
                    "reserved",
                ],
            },
            "uniqueItems": True,
        },
        "suspendprogram": {"type": "string"},
        "suspendrate": {"type": "integer", "minimum": 0},
        "suspendtime": {
            "oneOf": [{"type": "integer", "minimum": -1}, {"type": "string", "enum": ["infinite"]}]
        },
        "suspendtimeout": {"type": "integer", "minimum": 0},
        "switchparameters": {
            "type": "object",
            "properties": {
                "vnis": {"type": "string"},
                "tcs": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "enum": [
                            "dedicated_access",
                            "low_latency",
                            "bulk_data",
                            "best_effort",
                            "tc_dedicated_access",
                            "tc_low_latency",
                            "tc_bulk_data",
                            "tc_best_effort",
                        ],
                    },
                    "uniqueItems": True,
                },
                "single_node_vni": {"type": "string", "enum": ["all", "user", "none"]},
                "job_vni": {"type": "string", "enum": ["all", "user", "none"]},
                "adjust_limits": {"type": "boolean"},
                "no_adjust_limits": {"type": "boolean"},
                "hwcoll_addrs_per_job": {"type": "boolean"},
                "hwcoll_num_nodes": {"type": "boolean"},
                "fm_url": {"type": "boolean"},
                "fm_auth": {"type": "string", "enum": ["oauth", "basic"]},
                "fm_authdir": {"type": "string"},
                "def_txqs": {"type": "integer", "minimum": 0},
                "res_txqs": {"type": "integer", "minimum": 0, "maximum": 1024},
                "max_txqs": {"type": "integer", "minimum": 0, "maximum": 1024},
                "def_tgqs": {"type": "integer", "minimum": 0},
                "res_tgqs": {"type": "integer", "minimum": 0, "maximum": 512},
                "max_tgqs": {"type": "integer", "minimum": 0, "maximum": 512},
                "def_eqs": {"type": "integer", "minimum": 0},
                "res_eqs": {"type": "integer", "minimum": 0, "maximum": 2047},
                "max_eqs": {"type": "integer", "minimum": 0, "maximum": 2047},
                "def_cts": {"type": "integer", "minimum": 0},
                "res_cts": {"type": "integer", "minimum": 0, "maximum": 2047},
                "max_cts": {"type": "integer", "minimum": 0, "maximum": 2047},
                "def_tles": {"type": "integer", "minimum": 0},
                "res_tles": {"type": "integer", "minimum": 0, "maximum": 2048},
                "max_tles": {"type": "integer", "minimum": 0, "maximum": 2048},
                "def_ptes": {"type": "integer", "minimum": 0},
                "res_ptes": {"type": "integer", "minimum": 0, "maximum": 2048},
                "max_ptes": {"type": "integer", "minimum": 0, "maximum": 2048},
                "def_les": {"type": "integer", "minimum": 0},
                "res_les": {"type": "integer", "minimum": 0, "maximum": 16384},
                "max_les": {"type": "integer", "minimum": 0, "maximum": 16348},
                "def_acs": {"type": "integer", "minimum": 0},
                "res_acs": {"type": "integer", "minimum": 0, "maximum": 1022},
                "max_acs": {"type": "integer", "minimum": 0},
                "imex_channel_count": {"type": "integer", "minimum": 0},
            },
            "additionalProperties": False,
        },
        "switchtype": {"type": "string", "enum": ["switch/hpe_slingshot", "switch/nvidia_imex"]},
        "taskepilog": {"type": "string"},
        "taskplugin": {
            "type": "array",
            "items": {"type": "string", "enum": ["task/affinity", "task/cgroup"]},
            "uniqueItems": True,
        },
        "taskpluginparam": {
            "type": "object",
            "properties": {
                "cores": {"type": "boolean"},
                "none": {"type": "boolean"},
                "sockets": {"type": "boolean"},
                "threads": {"type": "boolean"},
                "slurmdoffspec": {"type": "boolean"},
                "verbose": {"type": "boolean"},
                "autobind": {"type": "string", "enum": ["cores", "sockets", "threads"]},
            },
            "additionalProperties": False,
        },
        "taskprolog": {"type": "string"},
        "tcptimeout": {"type": "integer", "minimum": 0},
        "tmpfs": {"type": "string"},
        "topologyparam": {
            "type": "object",
            "properties": {
                "dragonfly": {"type": "boolean"},
                "routepart": {"type": "boolean"},
                "switchasnoderank": {"type": "boolean"},
                "routetree": {"type": "boolean"},
                "topomaxsizeunroll": {"type": "integer", "minimum": -1},
                "topooptional": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "topologyplugin": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "topology/3d_torus",
                    "topology/block",
                    "topology/default",
                    "topology/tree",
                ],
            },
            "uniqueItems": True,
        },
        "trackwckey": {"type": "boolean"},
        "treewidth": {"type": "integer", "minimum": 0, "maximum": 65533},
        "unkillablestepprogram": {"type": "string"},
        "unkillablesteptimeout": {"type": "integer", "minimum": 0},
        "usepam": {"type": "boolean"},
        "vsizefactor": {"type": "integer", "minimum": 0, "maximum": 65533},
        "waittime": {"type": "integer", "minimum": 0, "maximum": 65533},
        "x11parameters": {
            "type": "object",
            "properties": {
                "home_xauthority": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "downnodes": DOWN_NODES_LIST_MODEL_SCHEMA,
        "frontendnodes": FRONTEND_NODE_MAPPING_MODEL_SCHEMA,
        "nodes": NODE_MAPPING_MODEL_SCHEMA,
        "nodesets": NODESET_MAPPING_MODEL_SCHEMA,
        "partitions": PARTITION_MAPPING_MODEL_SCHEMA,
    },
    "additionalProperties": False,
}


# `slurmdbd.conf` data model schema.

SLURMDBD_CONFIG_MODEL_SCHEMA = {
    "$schema": _GLOBAL_SCHEMA_VERSION,
    "type": "object",
    "properties": {
        "allownodefacct": {"type": "boolean"},
        "allresourcesabsolute": {"type": "boolean"},
        "archivedir": {"type": "string"},
        "archiveevents": {"type": "boolean"},
        "archivejobs": {"type": "boolean"},
        "archiveresvs": {"type": "boolean"},
        "archivescript": {"type": "string"},
        "archivesteps": {"type": "boolean"},
        "archivesuspend": {"type": "boolean"},
        "archivetxn": {"type": "boolean"},
        "archiveusage": {"type": "boolean"},
        "authalttypes": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["auth/jwt"],
            },
            "uniqueItems": True,
        },
        "authaltparameters": {
            "type": "object",
            "properties": {
                "jwks": {"type": "string"},
                "jwt_key": {"type": "string"},
            },
            "additionalProperties": False,
        },
        "authinfo": {
            "type": "object",
            "properties": {
                "socket": {"type": "string"},
                "ttl": {"type": "integer", "minimum": 0},
                "use_client_ids": {"type": "boolean"},
            },
            "additionalProperties": False,
        },
        "authtype": {
            "type": "string",
            "enum": ["auth/munge", "auth/slurm"],
        },
        "commitdelay": {"type": "integer", "minimum": 0},
        "communicationparameters": {
            "type": "object",
            "properties": {
                "disableipv4": {"type": "boolean"},
                "enableipv6": {"type": "boolean"},
                "keepaliveinterval": {"type": "integer", "minimum": 0},
                "keepaliveprobes": {"type": "integer", "minimum": 0},
                "keepalivetime": {"type": "integer", "minimum": 0},
            },
            "additionalProperties": False,
        },
        "dbdaddr": {"type": "string"},
        "dbdbackuphost": {"type": "string"},
        "dbdhost": {"type": "string"},
        "dbdport": {"type": "integer", "minimum": 0},
        "debugflags": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "auditrpcs",
                    "db_archive",
                    "db_assoc",
                    "db_event",
                    "db_job",
                    "db_qos",
                    "db_query",
                    "db_reservation",
                    "db_resource",
                    "db_step",
                    "db_tres",
                    "db_usage",
                    "db_wckey",
                    "federation",
                    "network",
                    "networkraw",
                    "tls",
                ],
            },
            "uniqueItems": True,
        },
        "debuglevel": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "debuglevelsyslog": {
            "type": "string",
            "enum": [
                "quiet",
                "fatal",
                "error",
                "info",
                "verbose",
                "debug",
                "debug2",
                "debug3",
                "debug4",
                "debug5",
            ],
        },
        "defaultqos": {"type": "string"},
        "disablecoorddbd": {"type": "boolean"},
        "hashplugin": {"type": "string", "enum": ["hash/k12", "hash/sha3"]},
        "logfile": {"type": "string"},
        "logtimeformat": {
            "type": "string",
            "enum": [
                "iso8601",
                "iso8601,format_stderr",
                "iso8601_ms",
                "iso8601_ms,format_stderr",
                "rfc5424",
                "rfc5424,format_stderr",
                "rfc5424_ms",
                "rfc5424_ms,format_stderr",
                "rfc3339",
                "rfc3339,format_stderr",
                "clock",
                "clock,format_stderr",
                "short",
                "short,format_stderr",
                "thread_id",
                "thread_id,format_stderr",
            ],
        },
        "maxquerytimerange": {"oneOf": [{"type": "integer", "minimum": 0}, {"type": "string"}]},
        "messagetimeout": {"type": "integer", "minimum": 0},
        "parameters": {
            "type": "object",
            "properties": {"preservecaseuser": {"type": "boolean"}},
            "additionalProperties": False,
        },
        "pidfile": {"type": "string"},
        "plugindir": {
            "type": "array",
            "items": {
                "type": "string",
            },
            "uniqueItems": True,
        },
        "privatedata": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": ["accounts", "events", "jobs", "reservations", "usage", "users"],
            },
            "uniqueItems": True,
        },
        "purgeeventafter": {"type": "string"},
        "purgejobafter": {"type": "string"},
        "purgeresvafter": {"type": "string"},
        "purgestepafter": {"type": "string"},
        "purgesuspendafter": {"type": "string"},
        "purgetxnafter": {"type": "string"},
        "purgeusageafter": {"type": "string"},
        "slurmuser": {"type": "string"},
        "storagebackuphost": {"type": "string"},
        "storagehost": {"type": "string"},
        "storageloc": {"type": "string"},
        "storageparameters": {
            "type": "object",
            "properties": {
                "ssl_cert": {"type": "string"},
                "ssl_ca": {"type": "string"},
                "ssl_capath": {"type": "string"},
                "ssl_key": {"type": "string"},
                "ssl_cipher": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                },
            },
            "additionalProperties": False,
        },
        "storagepass": {"type": "string"},
        "storageport": {"type": "integer", "minimum": 0},
        "storagetype": {"type": "string", "enum": ["accounting_storage/mysql"]},
        "storageuser": {"type": "string"},
        "tcptimeout": {"type": "integer", "minimum": 0},
        "trackslurmctlddown": {"type": "boolean"},
        "trackwckey": {"type": "boolean"},
    },
    "additionalProperties": False,
}



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base classes and functions for composing Slurm configuration editors."""

import os
import shutil
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Protocol, TypeVar

from .base import Model

_TModel = TypeVar("_TModel", bound=Model)


class BaseEditor(Protocol[_TModel]):
    """Base protocol for defining Slurm configuration file editors."""

    @property
    def __model__(self) -> type[_TModel]: ...  # noqa D105

    def dump(
        self,
        obj: _TModel,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None:
        """Atomically marshal a configuration model into a file."""
        # Write to a temp file then atomically replace the target file.
        target = Path(file)
        swap = target.with_stem("." + target.stem).with_suffix(target.suffix + ".swp")

        swap.write_text(self.dumps(obj) + "\n")
        _set_file_permissions(swap, mode=mode, user=user, group=group)
        swap.replace(target)

    def dumps(self, obj: _TModel, /) -> str:
        """Marshal a configuration model into a `str`."""
        return str(obj)

    def load(self, file: str | os.PathLike, /) -> _TModel:
        """Parse the contents of a file into a configuration model.

        Raises:
            FileNotFoundError: Raised if `file` does not exist.
        """
        return self.loads(Path(file).read_text())

    def loads(self, s: str, /) -> _TModel:
        """Parse a `str` into a configuration model."""
        return self.__model__.from_str(s)

    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[_TModel]:
        """Edit a configuration file.

        Args:
            file:
                Configuration file to edit.
                An empty model will be created if the file does not exist.
            mode: Access mode to set on the configuration file. (Default: rw-r--r--)
            user: User to set as owner of the configuration file. (Default: $USER)
            group: Group to set as owner of the configuration file. (Default: None)
        """
        if not os.path.exists(file):
            model = self.__model__()
        else:
            model = self.load(file)

        yield model
        self.dump(model, file, mode=mode, user=user, group=group)


def _set_file_permissions(
    file: str | os.PathLike,
    *,
    mode: int = 0o644,
    user: str | int | None = None,
    group: str | int | None = None,
) -> None:
    """Set permissions on a configuration file.

    Args:
        file: File to apply permission settings to.
        mode: Access mode to apply to file. (Default: rw-r--r--)
        user: User to set as owner of file. (Default: $USER)
        group: Group to set as owner of file. (Default: None)
    """
    Path(file).chmod(mode=mode)
    if user is None:
        user = os.getuid()

    shutil.chown(file, user, group)



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model representing the `slurmdbd.conf` configuration file."""

__all__ = ["SlurmdbdConfig", "SlurmdbdConfigEditor"]

from typing import Annotated, Any

from .core.base import Metadata, Model, classproperty
from .core.callback import (
    ColonSepCallback,
    CommaDictCallback,
    CommaSepCallback,
    StrBoolCallback,
)
from .core.editor import BaseEditor
from .core.schema import SLURMDBD_CONFIG_MODEL_SCHEMA


class SlurmdbdConfig(Model):
    """Model representing the `slurmdbd.conf` configuration file."""

    allow_no_def_acct: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    all_resources_absolute: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    archive_dir: str | None
    archive_events: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    archive_jobs: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    archive_resvs: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    archive_script: str | None
    archive_steps: Annotated[str | None, Metadata(callback=StrBoolCallback)]
    archive_suspend: Annotated[str | None, Metadata(callback=StrBoolCallback)]
    archive_txn: Annotated[str | None, Metadata(callback=StrBoolCallback)]
    archive_usage: Annotated[str | None, Metadata(callback=StrBoolCallback)]
    auth_alt_types: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    auth_alt_parameters: Annotated[dict[str, str] | None, Metadata(callback=CommaDictCallback)]
    auth_info: Annotated[dict[str, str | bool] | None, Metadata(callback=CommaDictCallback)]
    auth_type: str | None
    commit_delay: int | None
    communication_parameters: Annotated[
        dict[str, int | bool] | None,
        Metadata(callback=CommaDictCallback),
    ]
    dbd_addr: str | None
    dbd_backup_host: str | None
    dbd_host: str | None
    dbd_port: int | None
    debug_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    debug_level: str | None
    debug_level_syslog: str | None
    default_qos: str | None
    disable_coord_dbd: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    hash_plugin: str | None
    log_file: str | None
    log_time_format: str | None
    max_query_time_range: str | int | None
    message_timeout: str | None
    parameters: Annotated[dict[str, bool] | None, Metadata(callback=CommaDictCallback)]
    pid_file: str | None
    plugin_dir: Annotated[list[str] | None, Metadata(callback=ColonSepCallback)]
    private_data: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    purge_event_after: str | None
    purge_job_after: str | None
    purge_resv_after: str | None
    purge_step_after: str | None
    purge_suspend_after: str | None
    purge_txn_after: str | None
    purge_usage_after: str | None
    slurm_user: str | None
    storage_backup_host: str | None
    storage_host: str | None
    storage_loc: str | None
    storage_parameters: Annotated[
        dict[str, str | list[str]] | None,
        Metadata(callback=CommaDictCallback),
    ]
    storage_pass: str | None
    storage_port: int | None
    storage_type: str | None
    storage_user: str | None
    tcp_timeout: int | None
    track_slurmctld_down: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    track_wc_key: Annotated[bool | None, Metadata(callback=StrBoolCallback)]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return SLURMDBD_CONFIG_MODEL_SCHEMA


class SlurmdbdConfigEditor(BaseEditor):
    """Editor for the `slurmdbd.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return SlurmdbdConfig



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for streamlining common Slurm-related operations."""

__all__ = ["calculate_rs"]

from collections.abc import Iterable
from itertools import groupby


def calculate_rs(elements: Iterable[int]) -> str:
    """Calculate ranges and strides in an iterable.

    Args:
        elements: Iterable to calculate ranges and strides in.
            The iterable's elements must be unique and sortable in ascending order.

    Returns:
        A square-bracketed string with comma-separated ranges of consecutive values.
        example_input  = [0,1,2,3,4,5,6,8,9,10,12,14,15,16,18]
        example_output = '[0-6,8-10,12,14-16,18]'

    Notes:
        This function can be used to assist with converting arrays of resource names
        such as device files and node names into the Slurm hostname specification.
    """
    elements = sorted(elements)

    # The input is enumerate()-ed to produce a list of tuples of the elements and their indices.
    # groupby() uses the lambda key function to group these tuples by the difference between the
    # element and index. Consecutive values have equal difference between element and index,
    # so are grouped together. Hence, the elements of the first and last members of each
    # group give the range of consecutive values. If the group has only a single member, there are
    # no consecutive values either side of it (a "stride").
    out = "["
    for _, group in groupby(enumerate(elements), lambda elem: elem[1] - elem[0]):
        group = list(group)

        if len(group) == 1:
            # Single member, this is a stride.
            out += f"{group[0][1]},"
        else:
            # Range of consecutive values is first-last in group.
            out += f"{group[0][1]}-{group[-1][1]},"

    out = out.rstrip(",") + "]"
    return out



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base classes and functions for composing Slurm data models."""

# pyright: reportIncompatibleMethodOverride=false
#   This override is required because `pyright` dislikes in `ModelMapping` how
#   `_MapModel` and `MutableSequence` have conflicting signatures for the
#   `update` method. Potentially a candidate for a later refactor.

__all__ = [
    "classproperty",
    "Metadata",
    "Mode",
    "Model",
    "ModelList",
    "ModelMapping",
    "make_model_builder",
]

import json
from abc import ABC, ABCMeta, abstractmethod
from collections import Counter, deque
from collections.abc import (
    Callable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
)
from dataclasses import dataclass
from enum import Enum, auto
from functools import singledispatchmethod
from inspect import getmro
from itertools import filterfalse, takewhile
from types import MethodType, NoneType, UnionType
from typing import (
    Annotated,
    Any,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from jsonschema import ValidationError, validate
from typing_extensions import Self, get_original_bases

from ..exceptions import ModelError
from .callback import Callback, DefaultCallback


class classproperty(property):  # noqa N801
    """Class property attribute.

    Create a class-level, read-only property.
    """

    def __get__(self, __instance, __owner: Any | None = None) -> Any:  # noqa D105
        return super().__get__(__owner)


class subclassdispatch:  # noqa N801
    """Like `singledispatch`, but dispatch based on subclass instead.

    Dispatches functions based on what the passed argument is a subclass of.
    """

    def __init__(self, func: Callable[..., Any]) -> None:
        self._registry = {}
        self._default_func = func
        self.__name__ = func.__name__

    def register(self, func: Callable[..., Any]) -> None:
        _, cls = next(iter(get_type_hints(func).items()))
        self._registry[cls] = func

    def dispatch(self, sub_cls: Any) -> Callable[..., Any]:
        if not isinstance(sub_cls, type):
            sub_cls = type(sub_cls)

        for cls, func in self._registry.items():
            if issubclass(sub_cls, cls):
                return func

        return self._default_func

    def __call__(self, *args, **kwargs) -> Any:
        return self.dispatch(args[0])(*args, **kwargs)

    def __get__(self, instance, owner) -> Self | MethodType:
        if instance is not None:
            return MethodType(self, instance)
        else:
            return self


class Mode(Enum):
    r"""Mode to parse/marshal Slurm configuration data.

    Attributes:
        ONELINE: Parse/marshal Slurm configuration from/to one line.
        STANZA: Parse/marshal Slurm configuration from/to a stanza.

    Examples:
        If mode is set to `ONELINE`, parse/marshal using " " as the separator:
            parse:
                "key1=value key2=value" -> {"key1": "value", "key2": "value"}
            marshal:
                 {"key1": "value", "key2": "value"} -> "key1=value key2=value"

        If mode is set to `STANZA`, parse\\marshal using "\n" as the separator:
            parse:
                "key1=value\nkey2=value" -> {"key1": "value", "key2": "value"}
            marshal:
                {"key1": "value", "key2": "value"} -> "key1=value\nkey2=value"
    """

    ONELINE = auto()
    STANZA = auto()


@dataclass(frozen=True)
class Metadata:
    r"""Metadata for Slurm configuration field.

    Attributes:
        origin: Name of field in configuration model.
        alias: Alias(es) used to identify field in a configuration file.
        sep: Separator used to split a field's key and value.
        unique: Flag for if field can occur only once or multiple times in a configuration file.
        primary: Flag for if a field should be at the front of a line in marshalled output.
        callback: Callbacks for handling field during parsing and marshalling.
        default_factory:
            A zero argument callable that will provide a default value if either
            `del` is called on the field, or the field is not present in the configuration file.

    Methods:
        parse: Alias for accessing `callback.parser` method.
        marshal: Alias for accessing `callback.marshaller` method.
    """

    origin: str = ""
    alias: str = ""
    sep: str = "="
    unique: bool = True
    primary: bool = False
    callback: Callback = DefaultCallback
    default_factory: Callable[[], Any] | None = None

    def parse(self, expr: str) -> Any:
        """Alias for accessing `callback.parser` method."""
        return self.callback.parser(expr, self.sep)

    def marshal(self, key: str, value: str) -> str:
        """Alias for accessing `callback.marshaller` method."""
        return self.callback.marshaller(key, value, self.sep)


class _ModelMeta(ABCMeta):
    """Metaclass for defining configuration models."""

    def __new__(
        mcls,  # noqa N804
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        /,
        **kwargs: dict[str, Any],
    ) -> "_ModelMeta":
        field: str

        if "__annotations__" in namespace:
            for field, annotation in namespace["__annotations__"].items():
                metadata = _get_metadata(annotation)
                namespace[field] = _make_property(_format_field(field), metadata.default_factory)

        return super().__new__(mcls, name, bases, namespace, **kwargs)


class _ModelBase(metaclass=_ModelMeta):
    """Helper class for providing base methods to models through inheritance."""

    def __init__(self, instance: Any, /) -> None:
        instance = self.validate_model(instance)
        if self.__model_builder__:
            object.__setattr__(self, "_model_data", self.__model_builder__(instance))
        else:
            object.__setattr__(self, "_model_data", instance)

    @classproperty
    @abstractmethod
    def __model_schema__(self) -> dict[str, Any]:  # noqa N805
        """Model json schema."""

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None:  # noqa N805
        """Model builder."""
        return None

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N805
        """Model parse/marshal mode."""
        return Mode.STANZA

    @classmethod
    def from_str(cls, s: str, /) -> Self:
        """Create model from Slurm configuration string(s)."""
        return cls(_parse(cls, s))

    @classmethod
    def from_json(cls, obj: str | bytes | bytearray, /) -> Self:
        """Create model from a json object."""
        return cls(json.loads(obj))

    def json(self) -> str:
        """Get model as a json object."""
        return json.dumps(self.validate_model())

    def validate_model(self, instance: Iterable[Any] | None = None, /) -> Any:
        """Validate a model against its defined json schema.

        Args:
            instance:
                Optional dictionary object to validate against model json schema.
                If left set to `None`, this method will validate the attribute `_model_data`
                against the model json schema instead.

        Raises:
            ModelError:
                Raised if `instance` or the `_model_data` attribute fails to validate
                against the model's json schema.

        Returns:
            Expanded dictionary or list object. Sub-models must be rebuilt if validated output
            will be used for marshalling or overwriting the contents of the `_model_data`
            attribute. Output can be ignored if just performing a validation.
        """
        instance = (
            _get_as_builtin_object(instance)
            if instance is not None
            else _get_as_builtin_object(self._model_data)
        )
        try:
            validate(instance, schema=self.__model_schema__)
        except ValidationError as e:
            raise ModelError(e.message)

        return instance

    def __getattr__(self, item: str) -> Any:  # noqa D105
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def __setattr__(self, key: str, value: Any) -> Any:  # noqa D105
        if not hasattr(self, key):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

        super().__setattr__(key, value)

    def __str__(self) -> str:  # noqa D105
        self.validate_model()
        return _marshal(self)


_TModelBase = TypeVar("_TModelBase", bound=_ModelBase)
_KT = TypeVar("_KT", bound=str)


class _MapModel(_ModelBase, ABC):
    """Helper class for providing mapping-related methods to models through inheritance."""

    def __init__(self, m: dict[str, Any] | None = None, /, **kwargs) -> None:
        super().__init__((m if m else {}) | kwargs)

    @classmethod
    def from_dict(cls, m: dict[str, Any], /) -> Self:
        """Create model from dictionary object."""
        return cls(m)

    def dict(self) -> dict[str, Any]:
        """Get model as dictionary object."""
        return self.validate_model()

    def update(self, other: Self) -> None:
        """Update model content with content of another model."""
        self._model_data.update(other._model_data)


class Model(_MapModel, ABC):
    """Base class for build configuration models."""


class ModelList(_ModelBase, ABC, MutableSequence[_TModelBase]):
    """Base class for lists containing models."""

    def __init__(self, i: Model | Iterable[Model] | None = None, /, *models: Model) -> None:
        super().__init__(
            [i] + list(models) if isinstance(i, Model) else list(i or []) + list(models)
        )

    @classmethod
    def from_list(cls, i: Sequence[Any], /) -> Self:
        """Create model from list object."""
        return cls(i)

    def insert(self, index: int, value: Model) -> None:
        """Insert model before index."""
        self._model_data.insert(index, value)

    def list(self) -> list[Any]:
        """Get model as list object."""
        return self.validate_model()

    @singledispatchmethod
    def __getitem__(self, index) -> Model | Self:  # noqa D105
        raise TypeError(f"list indices must be integers or slices, not {type(index)}")

    @__getitem__.register
    def _(self, index: int) -> Model:
        return self._model_data[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(self._model_data[index])

    def __setitem__(self, index: int | slice, value: Model | Iterable[Model]) -> None:  # noqa D105
        self._model_data[index] = value

    def __delitem__(self, index: int | slice) -> None:  # noqa D105
        del self._model_data[index]

    def __len__(self) -> int:  # noqa D105
        return len(self._model_data)


def _sort_model_list(model_list: ModelList) -> dict[str, ModelList]:
    """Sort a model list by the wrapped model's primary key value."""
    result: dict[str, ModelList] = {}

    primary_key = _glom_primary_key(model_list)
    for model in model_list:
        primary_key_value = getattr(model, primary_key)
        if primary_key_value in result:
            result[primary_key_value].append(model)
        else:
            result[primary_key_value] = model_list.__class__(model)

    return result


class ModelMapping(_MapModel, ABC, MutableMapping[_KT, _TModelBase]):
    """Base class for mappings containing models."""

    def __getitem__(self, key: str, /) -> Any:  # noqa D105
        return self._model_data[key]

    def __setitem__(self, key: str, value: Any, /) -> None:  # noqa D105
        self._model_data[key] = value

    def __delitem__(self, key: str, /) -> None:  # noqa D105
        del self._model_data[key]

    def __iter__(self) -> Iterator[str]:  # noqa D105
        return iter(self._model_data)

    def __len__(self) -> int:  # noqa D105
        return len(self._model_data)


def make_model_builder(
    *models: type[Model] | type[ModelList] | type[ModelMapping],
) -> Callable[[Any], Any]:
    """Make a custom model builder from a provided list of models.

    Args:
        *models: Configuration model to create if json schema successfully matches.

    Returns:
        Callable for building a model from a base Python object.
    """

    def match(obj: Any) -> Model | ModelList | ModelMapping | None:
        for model in models:
            try:
                validate(obj, schema=cast(dict[str, Any], model.__model_schema__))
                return model(obj)
            except ValidationError:
                pass

        return None

    @subclassdispatch
    def build(obj: Any) -> Any:
        raise TypeError(f"expected list or Mapping, not {type(obj)}")

    @build.register
    def _(obj: Mapping) -> dict[str, Any]:
        result: dict[str, Any] = {}

        for k, v in obj.items():
            if model := match(v):
                built = model
            elif isinstance(v, Mapping | list):
                built = build(v)
            else:
                built = v

            result[k] = built

        return result

    @build.register
    def _(obj: list) -> list[Any]:
        result: list[Any] = []

        for v in obj:
            if model := match(v):
                built = model
            elif isinstance(v, Mapping | list):
                built = build(v)
            else:
                built = v

            result.append(built)

        return result

    return build


def _make_property(name: str, default_factory: Callable[[], Any] | None = None) -> property:
    """Make new property for a data model field.

    Args:
        name: Name of configuration option that a model field maps to.
        default_factory:
            Zero argument callable to invoke when `del` is either called on the model field
            or the model field is accessed but the configuration option is not present in the
            model's internal data map.
    """

    def getter(self: Model) -> Any | None:
        if default_factory and name not in self._model_data:
            self._model_data[name] = default_factory()

        return self._model_data.get(name, None)

    def setter(self: Model, v: Any) -> None:
        self._model_data[name] = v

    def deleter(self: Model) -> None:
        try:
            del self._model_data[name]
            if default_factory:
                self._model_data[name] = default_factory()
        except KeyError:
            pass

    return property(getter, setter, deleter)


def _make_annotation_map(
    model_t: _ModelBase, include_origin: bool = False
) -> dict[str, Annotated]:
    """Make a metadata mapping from a model class.

    Args:
        model_t: Model class to make metadata mapping from.
        include_origin: Include the metadata `origin` field in the annotation mapping.
    """
    annotation_map: dict[str, Annotated] = {}

    for field, annotation in get_type_hints(model_t, include_extras=True).items():
        metadata = _get_metadata(annotation)
        annotation_map[metadata.alias or _format_field(field)] = annotation
        if include_origin and metadata.origin:
            annotation_map[metadata.origin] = annotation

    return annotation_map


def _get_metadata(annotation: type) -> Metadata:
    """Get metadata for a Slurm configuration field.

    Args:
        annotation: Annotation to get metadata of.

    Notes:
        If the field is not annotated with `typing.Annotated`, a
        default `Metadata` object will be returned instead.
    """
    if hasattr(annotation, "__metadata__"):
        return annotation.__metadata__[0]

    return Metadata()


def _get_type(annotation: type) -> tuple[Any, ...]:
    """Get the expected type(s) of a Slurm configuration field.

    Args:
        annotation: Annotation to get the Slurm configuration field type from.

    Notes:
        Some Slurm configuration fields can be multiple types, so a tuple of types is
        returned rather than a single type. For example, the `slurm.conf` configuration
        options `MaxNodes` can be both an integer or the string "unlimited".
    """
    # Check if type annotation is wrapped with `typing.Annotated`.
    if hasattr(annotation, "__origin__"):
        annotation = annotation.__origin__

    # Check if origin is an "Optional". For example, is the annotation `str | None`?
    # Almost all annotations will be `<type> | None` since Slurm configuration
    # fields are nullable.
    if get_origin(annotation) is UnionType:
        return tuple(filterfalse(lambda t: t is NoneType, get_args(annotation)))
    else:
        return (annotation,)


def _get_primary_key(model_t: _ModelBase) -> str | None:
    r"""Get primary key of model.

    Args:
        model_t: Model to get primary key of.

    Returns:
        Model's primary key or `None` if the model has no declared primary key.
    """
    for field, annotation in get_type_hints(model_t, include_extras=True).items():
        metadata = _get_metadata(annotation)
        if metadata.primary:
            return field

    return None


def _glom_primary_key(model_list: ModelList) -> str:
    """Get the primary key of a list of models.

    Args:
        model_list: List of models to get the primary key of.

    Raises:
        ModelError: Raised if not all models in the list have the same primary key.
    """
    primary_keys = Counter([_get_primary_key(model) for model in model_list])
    if len(primary_keys) > 1 or None in primary_keys:
        raise ModelError(
            (
                f"models in {model_list} either have inconsistent primary keys"
                f"or do not have a primary key defined: {primary_keys}"
            )
        )

    return cast(str, list(primary_keys)[0])


def _format_field(field: str) -> str:
    """Format configuration field to be case-insensitive and remove underscores.

    Args:
        field: Configuration field to format.

    Examples:
        `_format_field("slurmctld_host")` -> "slurmctldhost"
    """
    return field.replace("_", "").lower()


@subclassdispatch
def _get_as_builtin_object(obj: Iterable[Any]) -> Iterable[Any]:
    """Recursively expand a complex iterable with nested models.

    Args:
        obj: Complex iterable to expand.
    """
    raise TypeError(f"expected list or Mapping, not {type(obj)}")


@_get_as_builtin_object.register
def _(obj: Mapping) -> dict[str, Any]:
    return {k: _expand_by_type(v) for k, v in obj.items()}


@_get_as_builtin_object.register
def _(obj: list) -> list[Any]:
    return [_expand_by_type(v) for v in obj]


def _expand_by_type(v: Any) -> Any:
    """Expand an object based on its type.

    Args:
        v: Object to expand based on its type.

    Returns:
        An "expanded" builtin object with no nested custom objects.
    """
    match v:
        case Model():
            return _get_as_builtin_object(v.dict())
        case ModelList():
            return _get_as_builtin_object(v.list())
        case Mapping() | list():
            return _get_as_builtin_object(v)
        case _:
            return v


@subclassdispatch
def _parse(model_t, s: str) -> Iterable[Any]:
    """Parse data in Slurm configuration format into models.

    Args:
        model_t: Model to create from parsed Slurm configuration data.
        s: Data in Slurm configuration format.

    Raises:
        TypeError: Raised if the type of `model_t` is not `Model`, `ModelList`, or `ModelMapping`.
        ModelError: Raised if an error occurs when parsing the value of `s`.
    """
    raise TypeError(
        f"expected subclass of type Model, ModelList, or ModelMapping, not {type(model_t)}"
    )


@_parse.register
def _(model_t: Model, s: str) -> dict[str, Any]:
    annotation_map = _make_annotation_map(model_t)

    classification_map: dict[str, list[str]] = {}
    for token, expr in _scan(s, mode=model_t.__model_mode__):
        if token not in annotation_map:
            raise ModelError(
                (
                    f"unrecognized configuration field `{token}`. "
                    f"valid configuration fields are: {list(annotation_map.keys())}"
                )
            )

        classification_map[token] = classification_map.get(token, []) + [expr]

    result: dict[str, Any] = {}
    for token, expr in classification_map.items():
        annotation = annotation_map[token]
        metadata = _get_metadata(annotation)
        tp = _get_type(annotation)

        if metadata.unique and len(expr) > 1:
            raise ModelError(
                (
                    f"expected configuration field `{token}` to be unique, "
                    f"but multiple entries found: {classification_map}"
                )
            )

        if _ModelBase in getmro(tp[0]):
            result[metadata.origin or token] = tp[0].from_str("\n".join(expr))
            continue

        if metadata.unique:
            result[metadata.origin or token] = metadata.parse(expr[0])
        else:
            result[metadata.origin or token] = [metadata.parse(e) for e in expr]

    return result


@_parse.register
def _(model_t: ModelList, s: str) -> list[Any]:
    result: list[Any] = []
    origin = get_original_bases(cast(type, model_t))[0]
    wrapped_model_t = get_args(origin)[0]
    for _, expr in _scan(s, model_t.__model_mode__):
        result.append(wrapped_model_t.from_str(expr))

    return result


@_parse.register
def _(model_t: ModelMapping, s: str) -> dict[str, Any]:
    result: dict[str, Any] = {}
    origin = get_original_bases(cast(type, model_t))[0]
    wrapped_model_t = get_args(origin)[1]
    primary_key = _get_primary_key(wrapped_model_t)

    if primary_key:
        for _, expr in _scan(s, model_t.__model_mode__):
            model = wrapped_model_t.from_str(expr)
            result.update({getattr(model, primary_key): model})
    elif primary_key is None and issubclass(wrapped_model_t, ModelList):
        model = wrapped_model_t.from_str(s)
        result.update(_sort_model_list(model))
    else:
        raise ModelError(
            (
                f"model {wrapped_model_t} has no primary key. "
                f"{wrapped_model_t} cannot be used with a model mapping"
            )
        )

    return result


def _clean(s: str) -> list[str]:
    """Clean configuration line(s) before further processing.

    Args:
        s: Configuration line(s) to clean.

    Warnings:
        This function will strip all comments from the configuration line(s),
        but will not preserve them. Original comments will not be present in
        the marshalled configuration data output.
    """
    result: list[str] = []

    for line in s.splitlines():
        cleaned = line.split("#", maxsplit=1)[0].strip()
        if cleaned == "":  # Line was a comment line or is empty.
            continue

        result.append(cleaned)

    return result


def _scan(s: str, /, mode: Mode) -> Iterable[tuple[str, str]]:
    """Scan Slurm configuration data to identify key-value pairs.

    Args:
        s: Configuration data to scan for key-value pairs.
        mode: Mode to scan file in.
    """
    match mode:
        case mode.ONELINE:
            expr = _scan_line(c[0]) if (c := _clean(s)) else []

        case mode.STANZA:
            expr = _clean(s)

        case _:
            raise TypeError(f"mode must be either ONELINE or STANZA, not {type(mode)}")

    for e in expr:
        token = "".join(map(str, takewhile(lambda c: c != "=" and c != " ", e))).lower()
        yield token, e


def _scan_line(s: str) -> list[str]:
    """Scan a Slurm configuration line to identify key-value pairs.

    Args:
        s: Configuration line to scan for key-value pairs.

    Notes:
        This method is used rather than `shlex.split(...)` or a regular expression
        because this method offers a ~429% performance increase over `shlex.split(...)`
        for identifying key-value pairs in a Slurm configuration line.
    """
    pairs: list[str] = []
    quote = None
    current_pair = ""

    for c in s:
        if c in ['"', "'"]:
            if c == quote:
                quote = None
                current_pair += c
            elif quote is None:
                quote = c
                current_pair += c
            else:
                current_pair += c
        elif c.isspace() and quote is None:
            if current_pair:
                pairs.append(current_pair)
                current_pair = ""
        else:
            current_pair += c

    if current_pair:
        pairs.append(current_pair)

    return pairs


@subclassdispatch
def _marshal(model) -> str:
    """Marshal a model into Slurm configuration format.

    Args:
        model: Configuration model to marshal into Slurm configuration format.
    """
    raise TypeError(
        f"expected subclass of type Model, ModelList, or ModelMapping, not {type(model)}"
    )


@_marshal.register
def _(model: Model) -> str:
    annotation_map = _make_annotation_map(model, include_origin=True)
    result: deque[str] = deque()

    # Directly access `_model_data` here because `_expand(...)` does not
    # preserve the type information required for subclass dispatching.
    for k, v in model._model_data.items():  # noqa SLF001
        if isinstance(v, Model | ModelList | ModelMapping):
            result.append(str(v))
            continue

        metadata = _get_metadata(annotation_map[k])
        expr = metadata.marshal(metadata.alias or k, v)
        if metadata.primary:
            result.appendleft(expr)
        else:
            result.append(expr)

    return _format_output(result, mode=model.__model_mode__)


@_marshal.register
def _(model: ModelList) -> str:
    result: list[str] = []
    for wrapped_model in model:
        result.append(str(wrapped_model))

    return _format_output(result, mode=model.__model_mode__)


@_marshal.register
def _(model: ModelMapping) -> str:
    result: list[str] = []
    for wrapped_model in model.values():
        result.append(str(wrapped_model))

    return _format_output(result, mode=model.__model_mode__)


def _format_output(i: Iterable[str], mode: Mode) -> str:
    """Format `_marshal` output based on model mode.

    Args:
        i: Iterable to format based on given mode.
        mode: Mode to format output in.

    Raises:
        TypeError:
            Raised if invalid marshaling mode is provided.
            Valid modes are `ONELINE` and `STANZA`.
    """
    match mode:
        case Mode.ONELINE:
            return " ".join(i)
        case Mode.STANZA:
            return "\n".join(i)
        case _:
            raise TypeError(f"mode must be either ONELINE or STANZA, not {type(mode)}")



#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for models and editor of the `gres.conf` configuration file."""

from constants import EXAMPLE_GRES_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import Gres, GresConfig, ModelError, cgroupconfig, gresconfig

EXPECTED_GRES_DUMPS_OUTPUT = """
autodetect=nvml
name=gpu type=gp100 file=/dev/nvidia0 cores=0,1
name=gpu type=gp100 file=/dev/nvidia1 cores=0,1
name=gpu type=p6000 file=/dev/nvidia2 cores=2,3
name=gpu type=p6000 file=/dev/nvidia3 cores=2,3
name=gpu nodename=juju-abc654-1 type=tesla_t4 file=/dev/nvidia[0-1] count=8G
name=gpu nodename=juju-abc654-1 type=l40s file=/dev/nvidia[2-3] count=12G
name=mps count=200 file=/dev/nvidia0
name=mps count=200 file=/dev/nvidia1
name=mps count=100 file=/dev/nvidia2
name=mps count=100 file=/dev/nvidia3
name=bandwidth type=lustre count=4G flags=countonly
""".strip()


class TestGresConfig(TestCase):
    """Unit tests for models and the editor of the `gres.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/gres.conf", contents=EXAMPLE_GRES_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of `gres.conf` model editor."""
        config = gresconfig.loads(EXAMPLE_GRES_CONFIG)
        self.assertEqual(config.auto_detect, "nvml")
        self.assertDictEqual(
            config.gres.dict(),
            {
                "gpu": [
                    {"name": "gpu", "type": "gp100", "file": "/dev/nvidia0", "cores": [0, 1]},
                    {"name": "gpu", "type": "gp100", "file": "/dev/nvidia1", "cores": [0, 1]},
                    {"name": "gpu", "type": "p6000", "file": "/dev/nvidia2", "cores": [2, 3]},
                    {"name": "gpu", "type": "p6000", "file": "/dev/nvidia3", "cores": [2, 3]},
                    {
                        "name": "gpu",
                        "nodename": "juju-abc654-1",
                        "type": "tesla_t4",
                        "file": "/dev/nvidia[0-1]",
                        "count": "8G",
                    },
                    {
                        "name": "gpu",
                        "nodename": "juju-abc654-1",
                        "type": "l40s",
                        "file": "/dev/nvidia[2-3]",
                        "count": "12G",
                    },
                ],
                "mps": [
                    {"name": "mps", "count": 200, "file": "/dev/nvidia0"},
                    {"name": "mps", "count": 200, "file": "/dev/nvidia1"},
                    {"name": "mps", "count": 100, "file": "/dev/nvidia2"},
                    {"name": "mps", "count": 100, "file": "/dev/nvidia3"},
                ],
                "bandwidth": [
                    {"name": "bandwidth", "type": "lustre", "count": "4G", "flags": ["countonly"]}
                ],
            },
        )

    def test_dumps(self) -> None:
        """Test `dumps` method of the `gres.conf` model editor."""
        config = gresconfig.loads(EXAMPLE_GRES_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(cgroupconfig.dumps(config), EXPECTED_GRES_DUMPS_OUTPUT)
        # New and old config should not be equal since editor strips all comments.
        self.assertNotEqual(cgroupconfig.dumps(config), EXAMPLE_GRES_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` method of the `gres.conf` model editor."""
        with gresconfig.edit("/etc/slurm/gres.conf") as config:
            config.auto_detect = "rsmi"
            config.gres["gpu"].append(Gres(name="gpu", type="epyc", file="/dev/amd0"))

        config = gresconfig.load("/etc/slurm/gres.conf")
        self.assertEqual(config.auto_detect, "rsmi")
        self.assertDictEqual(
            config.gres["gpu"][-1].dict(), {"name": "gpu", "type": "epyc", "file": "/dev/amd0"}
        )

    def test_load_fail(self) -> None:
        """Test that bogus values are automatically caught when loading a new model."""
        # Catch if field is unexpected or in the incorrect syntax.
        with self.assertRaises(ModelError):
            GresConfig.from_str(
                """
                Accelerator=EXPENSIVE_GPU AutoDetect=False
                """
            )

        # Catch if field value is invalid when checking against the model schema.
        with self.assertRaises(ModelError):
            GresConfig.from_str(
                """
                autodetect=superpowerfulgpu
                name=gpu file=65
                """
            )

        with self.assertRaises(ModelError):
            Gres.from_str(
                """
                name=gpu type=gp100 file=/dev/nvidia0 flags=countonly,custom_gpu_env
                """
            )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = gresconfig.loads(EXAMPLE_GRES_CONFIG)
            config.auto = "nvidia"

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = gresconfig.loads(EXAMPLE_GRES_CONFIG)
            _ = config.auto
